"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from pcraster._pcraster import Scalar

from zusatz import *
from netCDF4 import num2date, date2num, default_fillvals
from pandas import date_range
import numpy as np
import time as xtime
import os
import globals
from bisect import bisect_left

# ------------------------------

def defsoil(name1, name2=None, name3=None):
    """ loads 3 array in a list
    """
    try:
        in1 = loadmap(name1)
    except:
        in1 = name1
    if name2 is None:
        in2 = in1
    else:
        try:
            in2 = loadmap(name2)
        except:
            in2 = name2
    if name3 is None:
        in3 = in1
    else:
        try:
            in3 = loadmap(name3)
        except:
            in3 = name3
    return [in1, in2, in3]

def valuecell(mask, coordx, coordstr):
    """
    to put a value into a pcraster map -> invert of cellvalue
    pcraster map is converted into a numpy array first
    """
    coord = []
    for xy in coordx:
        try:
            coord.append(float(xy))
        except:
            msg = "Gauges: " + xy + " in " + coordstr + " is not a coordinate"
            raise LisfloodError(msg)

    null = np.zeros((pcraster.clone().nrRows(), pcraster.clone().nrCols()))
    null[null == 0] = -9999

    for i in xrange(int(len(coord) / 2)):
        col = int(
            (coord[i * 2] - pcraster.clone().west()) / pcraster.clone().cellSize())
        row = int(
            (pcraster.clone().north() - coord[i * 2 + 1]) / pcraster.clone().cellSize())
        #if col >= 0 and row >= 0 and col < pcraster.clone().nrCols() and row < pcraster.clone().nrRows():
        if col >= 0 and row >= 0 and col < pcraster.clone().nrCols() and row < pcraster.clone().nrRows():
            null[row, col] = i + 1
        else:
            msg = "Coordinates: " + str(coord[i * 2]) + ',' + str(
                coord[i * 2 + 1]) + " to put value in is outside mask map - col,row: " + str(col) + ',' + str(row)
            raise LisfloodError(msg)

    map = numpy2pcr(Nominal, null, -9999)
    return map


def metaNetCDF():
    """ Get metadata information from netcdf map.
    
    metaNetCDF gets metadata information from netCDF template file. Name and path of template file are stored
    in settings.xml file using key "netCDFtemplate". If template file is not available, information is read
    from "E0Maps".
    
    :return: metadataNCDF as global dictionary
    """
    try:
        # using ncdftemplate
        filename = os.path.splitext(binding['netCDFtemplate'])[0] + '.nc'
        nf1 = iterOpenNetcdf(filename, "", 'r')
        for var in nf1.variables:
           metadataNCDF[var] = {k: v for k, v in nf1.variables[var].__dict__.iteritems() if k !='_FillValue'}
        nf1.close()
        return
    except:
        pass
    # if no template .nc is given the e.nc file is used
    filename = os.path.splitext(binding['E0Maps'])[0] + '.nc'
    nf1 = iterOpenNetcdf(filename, "Trying to get metadata from netcdf \n", 'r')
    for var in nf1.variables:
       metadataNCDF[var] = {k: v for k, v in nf1.variables[var].__dict__.iteritems() if k !='_FillValue'}
    nf1.close()


def mapattrNetCDF(name):
    """
    get the map attributes like col, row etc from a ntcdf map
    and define the rectangular of the mask map inside the netcdf map
    """
    filename = os.path.splitext(name)[0] + '.nc'
    nf1 = iterOpenNetcdf(filename, "Checking netcdf map \n", 'r')
    spatial_dims = ('x', 'y') if 'x' in nf1.variables else ('lon', 'lat')
    x1, x2, y1, y2 = [round(nf1.variables[v][j], 5) for v in spatial_dims for j in (0, 1)]
    nf1.close()
    if maskmapAttr['cell'] != round(np.abs(x2 - x1), 5) or maskmapAttr['cell'] != round(np.abs(y2 - y1), 5):
        raise LisfloodError("Cell size different in maskmap {} and {}".format(binding['MaskMap'], filename))
    half_cell = maskmapAttr['cell'] / 2
    x = x1 - half_cell # |
    y = y1 + half_cell # | coordinates of the upper left corner of the input file upper left pixel
    cut0 = int(round(np.abs(maskmapAttr['x'] - x) / maskmapAttr['cell']))
    cut1 = cut0 + maskmapAttr['col']
    cut2 = int(round(np.abs(maskmapAttr['y'] - y) / maskmapAttr['cell']))
    cut3 = cut2 + maskmapAttr['row']
    return cut0, cut1, cut2, cut3 # input data will be sliced using [cut0:cut1,cut2:cut3]


def loadsetclone(name):
    """ Load 'MaskMap' and set as clone
        
    :param name: name of the key in Settings.xml containing path and name of mask map as string
    :return: map: mask map (False=include in modelling; True=exclude from modelling) as pcraster
    """
    filename = os.path.normpath(binding[name])
    if not os.path.exists(filename):
        raise LisfloodError('File not existing: {}'.format(filename))
    coord = filename.split()    #returns a list of all the words in the string
    if len(coord) == 5:
        # changed order of x, y i- in setclone y is first in Lisflood
        # settings x is first
        # setclone row col cellsize xupleft yupleft
        try:
            setclone(int(coord[1]), int(coord[0]), float(coord[2]), float(coord[3]), float(coord[4]))   # CM: pcraster
        except:
            rem = "["+str(coord[0])+" "+ str(coord[1])+" "+ str(coord[2])+" "+ str(coord[3])+" "+str(coord[4])+"]"
            msg = "Maskmap: " + rem + \
                  " are not valid coordinates (col row cellsize xupleft yupleft)"
            raise LisfloodError(msg)
        mapnp = np.ones((int(coord[1]), int(coord[0])))
        #mapnp[mapnp == 0] = 1
        map = numpy2pcr(Boolean, mapnp, -9999)
    elif len(coord) == 1:
        #read information on clone map from map (pcraster or netcdf)
        try:
            # try to read a pcraster map
            iterSetClonePCR(filename)
            map = boolean(iterReadPCRasterMap(filename))
            flagmap = True
            mapnp = pcr2numpy(map,np.nan)

        except Exception as e:
            print(type(e))
            print(str(e))
            # try to read a netcdf file
            filename = os.path.splitext(binding[name])[0] + '.nc'
            nf1 = iterOpenNetcdf(filename, "", "r")
            value = nf1.variables.items()[-1][0]  # get the last variable name

            if 'x' in nf1.variables:
                x1 = nf1.variables['x'][0]
                x2 = nf1.variables['x'][1]
                xlast = nf1.variables['x'][-1]
                y1 = nf1.variables['y'][0]
                ylast = nf1.variables['y'][-1]
            else:
                x1 = nf1.variables['lon'][0]
                x2 = nf1.variables['lon'][1]
                y1 = nf1.variables['lat'][0]
                y2 = nf1.variables['lat'][1]
                xlast = nf1.variables['lat'][-1]
                ylast = nf1.variables['lon'][-1]

            cellSize = round(np.abs(x2 - x1),4)
            nrRows = int(0.5+np.abs(ylast - y1) / cellSize + 1)
            nrCols = int(0.5+np.abs(xlast - x1) / cellSize + 1)
            x = x1 - cellSize / 2
            y = y1 + cellSize / 2
            mapnp = np.array(nf1.variables[value][0:nrRows, 0:nrCols])
            nf1.close()
            # setclone  row col cellsize xupleft yupleft
            setclone(nrRows,nrCols, cellSize, x, y)
            map = numpy2pcr(Boolean, mapnp, 0)
            #map = boolean(map)
            flagmap = True
        if Flags['check']:
            checkmap(name, filename, map, flagmap, 0)
    else:
        msg = "Maskmap: " + Mask + \
            " is not a valid mask map nor valid coordinates"
        raise LisfloodError(msg)

    # Definition of cellsize, coordinates of the meteomaps and maskmap
    # need some love for error handling
    maskmapAttr['x'] = pcraster.clone().west()          # mask map West bound
    maskmapAttr['y'] = pcraster.clone().north()         # mask map North bound
    maskmapAttr['col'] = pcraster.clone().nrCols()      # mask map number of columns
    maskmapAttr['row'] = pcraster.clone().nrRows()      # mask map number of rows
    maskmapAttr['cell'] = pcraster.clone().cellSize()   # mask map cell size

    # put in the ldd map
    # if there is no ldd at a cell, this cell should be excluded from modelling
    ldd = loadmap('Ldd',pcr=True)
    #convert ldd to numpy
    maskldd = pcr2numpy(ldd,np.nan)
    #convert numpy map to 8bit
    maskarea = np.bool8(mapnp)
    #compute mask (pixels in maskldd AND maskarea)
    mask = np.logical_not(np.logical_and(maskldd,maskarea))


    #return all the non-masked data as a 1-D array
    mapC = np.ma.compressed(np.ma.masked_array(mask,mask))
    # Definition of compressed array and info how to blow it up again
    maskinfo['mask']=mask
    maskinfo['shape']=mask.shape
    maskinfo['maskflat']=mask.ravel()    # map to 1D not compresses
    maskinfo['shapeflat']=maskinfo['maskflat'].shape   #length of the 1D array
    maskinfo['mapC']=mapC.shape                        # length of the compressed 1D array
    maskinfo['maskall'] =np.ma.masked_all(maskinfo['shapeflat'])  # empty map 1D but with mask
    maskinfo['maskall'].mask = maskinfo['maskflat']
    globals.inZero=np.zeros(maskinfo['mapC'])
    return map


def compressArray(map,pcr=True,name="None"):
        if pcr:
            mapnp = pcr2numpy(map,np.nan)
            mapnp1 = np.ma.masked_array(mapnp,maskinfo['mask'])
        else:
            mapnp1 = np.ma.masked_array(map,maskinfo['mask'])
        mapC = np.ma.compressed(mapnp1)
        #if fill: mapC[np.isnan(mapC)]=0
        if name != "None":
            if np.max(np.isnan(mapC)):
                msg = name + " has less valid pixels than area or ldd \n"
                raise LisfloodError(msg)
                # test if map has less valid pixel than area.map (or ldd)
        return mapC.astype(float)

def decompress(map):
     #dmap=np.ma.masked_all(maskinfo['shapeflat'], dtype=map.dtype)
     dmap=maskinfo['maskall'].copy()
     dmap[~maskinfo['maskflat']] = map[:]
     dmap = dmap.reshape(maskinfo['shape'])
     # check if integer map (like outlets, lakes etc
     try:
        checkint=str(map.dtype)
     except:
        checkint="x"
     if checkint in ["int16", "int32", "int64"]:
          dmap[dmap.mask]=-9999
          map = numpy2pcr(Nominal, dmap, -9999)
     elif checkint=="int8":
          dmap[dmap<0]=-9999
          map = numpy2pcr(Nominal, dmap, -9999)
     else:
          dmap[dmap.mask]=-9999
          map = numpy2pcr(Scalar, dmap, -9999)
     return map


def makenumpy(map):
    if not('numpy.ndarray' in str(type(map))):
        out = np.empty(maskinfo['mapC'])
        out.fill(map)
        return out
    else: return map


def loadmap(name,pcr=False, lddflag=False, timestampflag='exact', averageyearflag=False):
    """ Load a static map either value or pcraster map or netcdf (single or stack)
    
    Load a static map either value or pcraster map or netcdf (single or stack)
    If a netCDF stack is loaded, map is read according to timestepInit date (i.e. model time step). If timestepInit is a
    step number, step number is converted to date (referred to CalendarDayStart in settings.xml). Then date is used to
    read time step from netCDF file.
    if timestampflag = 'closest' and loadmap is reading a NetCDF stack, the timestep with the closest timestamp will be
    loaded if the exact one is not available.
    
    :param name: name of key in Settings.xml input file containing path and name of the map file (as string)
    :param pcr: flag for output maps in pcraster format 
    :param lddflag: flag for local drain direction map (CM??)
    :param timestampflag: look for exact time stamp in netcdf file ('exact') or for the closest (left) time stamp available ('closest')
    :param averageyearflag: if True, use "average year" netcdf file over the entire model simulation period
    :return: map or mapC
    :except: pcr: maps must have the same size of clone.map
             netCDF: time step timestepInit must be included into the stack 
    """

    #name of the key in Settimgs.xml file containing path and name of the map file
    value = binding[name]
    #path and name of the map file
    filename = value
    load = False
    pcrmap = False
    #try reading in PCRaster map format
    try:
        #try reading pcraster map
        mapC = float(value)
        flagmap = False
        load = True
        if pcr: map=mapC
    except ValueError:
        try:
            #try reading pcraster map exploiting the iterAccess class
            map = iterReadPCRasterMap(value)
            flagmap = True
            load = True
            pcrmap = True
        except:
            load = False
    if load and pcrmap:
        #map is loaded and it is in pcraster format
        try:
           test = pcraster.scalar(map) + pcraster.scalar(map)    # test if map is same size as clone map, if not it will make an error
        except:
           msg = value +" might be of a different size than clone size "
           raise LisfloodError(msg)
    #if failed before try reading from netCDF map format
    if not(load):
        # read a netcdf  (single one not a stack)
        filename = os.path.splitext(value)[0] + '.nc'
        # get mapextend of netcdf map and calculate the cutting
        cut0, cut1, cut2, cut3 = mapattrNetCDF(filename)
        # load netcdf map but only the rectangle needed
        nf1 = iterOpenNetcdf(filename, "", 'r')
        value = nf1.variables.items()[-1][0]
        # get the last variable name (it must be the variable to be read by Lisflood)
        if not timestepInit:
            #if timestepInit is missing, read netcdf as single static map
            mapnp = nf1.variables[value][cut2:cut3, cut0:cut1]
        else:
            if 'time' in nf1.variables:
                #read a netcdf  (stack) - state files
                #get information from netCDF stack
                t_steps = nf1.variables['time'][:]  # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
                t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
                t_cal = getCalendarType(nf1)
                #get year from time unit in case average year is used
                if averageyearflag:
                    #get date of the first step in netCDF file containing average year values
                    first_date = num2date(t_steps[0],t_unit,t_cal)
                    #get year of the first step in netCDF file containing average year values
                    t_ref_year = first_date.year

                #select timestep to use for reading from netCDF stack based on timestepInit (state file time step)
                timestepI = Calendar(timestepInit[0])
                if type(timestepI) is datetime.datetime:
                    #reading dates in XML settings file
                    #get step id number in netCDF stack for timestepInit date
                    if averageyearflag:
                        #if using an average year don't care about the year in timestepIDate and change it to the netCDF first time step year
                        try:
                            timestepI = timestepI.replace(year=t_ref_year)
                        except:
                            timestepI = timestepI.replace(day=28)
                            timestepI = timestepI.replace(year=t_ref_year)
                    timestepI = date2num(timestepI,nf1.variables['time'].units)
                else:
                    #reading step numbers in XML file
                    #timestepI = int(timestepI) -1
                    begin = Calendar(binding['CalendarDayStart'])
                    DtSec = float(binding['DtSec'])
                    DtDay = float(DtSec / 86400)
                    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)
                    # get date for step number timestepI (referred to CalendarDayStart)
                    timestepIDate = begin + datetime.timedelta(days=(timestepI - 1) * DtDay)
                    #get step id number in netCDF stack for step timestepInit
                    #timestepInit refers to CalenradDayStart
                    #timestepI now refers to first date in netCDF stack
                    if averageyearflag:
                        #using an average year, don't care about the year in timestepIDate and change it to the netCDF time unit year
                        try:
                            timestepIDate = timestepIDate.replace(year=t_ref_year)
                        except:
                            #if simulation year is leap and average year is not, switch 29/2 with 28/2
                            timestepIDate = timestepIDate.replace(day=28)
                            timestepIDate = timestepIDate.replace(year=t_ref_year)
                    timestepI = date2num(timestepIDate,units = t_unit, calendar = t_cal)

                if not(timestepI in nf1.variables['time'][:]):
                    if (timestampflag == 'exact'):
                        #look for exact time stamp when loading data
                        msg = "time step " + str(int(timestepI)+1)+" is not stored in "+ filename
                        raise LisfloodError(msg)
                    elif (timestampflag == 'closest'):
                        #get the closest value
                        timestepInew = takeClosest(t_steps, timestepI)
                        #set timestepI to the closest available time step in netCDF file
                        timestepI = timestepInew

                itime = np.where(nf1.variables['time'][:] == timestepI)[0][0]
                mapnp = nf1.variables[value][itime,cut2:cut3, cut0:cut1]
            else:
                # read a netcdf  (single one)
                mapnp = nf1.variables[value][cut2:cut3, cut0:cut1]

        #masking
        try:
            if any(maskinfo):
                mapnp.mask = maskinfo['mask']
        except:
            pass
        nf1.close()
        # if a map should be pcraster
        if pcr:
            # check if integer map (like outlets, lakes etc
            checkint=str(mapnp.dtype)
            if checkint=="int16" or checkint=="int32":
                mapnp[mapnp.mask]=-9999
                map = numpy2pcr(Nominal, mapnp, -9999)
            elif checkint=="int8":
                mapnp[mapnp<0]=-9999
                map = numpy2pcr(Nominal, mapnp, -9999)
            else:
                mapnp[np.isnan(mapnp)] = -9999
                map = numpy2pcr(Scalar, mapnp, -9999)
            # if the map is a ldd
            #if value.split('.')[0][:3] == 'ldd':
            if lddflag: map = ldd(nominal(map))
        else:
            mapC = compressArray(mapnp,pcr=False,name=filename)
        flagmap = True
    # pcraster map but it has to be an array
    if pcrmap and not(pcr):
        mapC = compressArray(map,name=filename)
    if Flags['check']:
        print name, filename
        if flagmap == False:
            checkmap(name, filename, mapC, flagmap, 0)
        elif pcr:
            checkmap(name, filename, map, flagmap, 0)
        else:
            print name, mapC.size
            if mapC.size >0:
                map= decompress(mapC)
                checkmap(name, filename, map, flagmap, 0)
    if pcr:
        return map
    elif isinstance(mapC, np.ndarray):
        return mapC.astype(float)
    else:
        return mapC



def takeClosest(myList, myNumber):
    """ Returns the closest left value to myNumber in myList
    
    Assumes myList is sorted. Returns closest left value to myNumber.
    If myList is sorted in raising order, it returns the closest smallest value.
    https://stackoverflow.com/questions/12141150/from-list-of-integers-get-number-closest-to-a-given-value
    
    :param myList: list of ordered values
    :param myNumber: number to be searche in myList
    :return: closest left number to myNumber in myList
    """
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    # after = myList[pos]
    # if after - myNumber < myNumber - before:
    #    return after
    # else:
    return before




def loadLAI(value, pcrvalue, i,pcr=False):
    """
    load Leaf are map stacks  or water use maps stacks
    """
    pcrmap = False
    try:
        map = iterReadPCRasterMap(pcrvalue)
        filename = pcrvalue
        pcrmap = True
    except:
        filename = os.path.splitext(value)[0] + '.nc'
        # get mapextend of netcdf map
        # and calculate the cutting
        cut0, cut1, cut2, cut3 = mapattrNetCDF(filename)
        nf1 = iterOpenNetcdf(filename, "", 'r')
        value = nf1.variables.items()[-1][0]  # get the last variable name
        mapnp = nf1.variables[value][i, cut2:cut3, cut0:cut1]
        nf1.close()
        mapC = compressArray(mapnp,pcr=False,name=filename)
        # mapnp[np.isnan(mapnp)] = -9999
        # map = numpy2pcr(Scalar, mapnp, -9999)
        # if check use a pcraster map
        if Flags['check' or pcr]:
            map = decompress(mapC)
    if pcrmap: mapC = compressArray(map,name=filename)
    if Flags['check']:
        checkmap(os.path.basename(pcrvalue), filename, map, True, 0)
    if pcr: return map
    return mapC


def readmapsparse(name, time, oldmap):
    """
    load stack of maps 1 at each timestamp in Pcraster format
    """
    filename = generateName(name, time)
    try:
        map = iterReadPCRasterMap(filename)
        find = 1
    except:
        find = 2
        if oldmap is None:
            for i in range(time - 1, 0, -1):
                altfilename = generateName(name, i)
                if os.path.exists(altfilename):
                    map = iterReadPCRasterMap(altfilename)
                    find = 1
                    # break
            if find == 2:
                msg = "no map in stack has a smaller time stamp than: " + filename
                raise LisfloodError(msg)
        else:
            map = oldmap
            if Flags['loud']:
                s = " last_%s" % (os.path.basename(name))
                print s,
    if Flags['check']:
        checkmap(os.path.basename(name), filename, map, True, find)
    mapC = compressArray(map,name=filename)
    return mapC


def readnetcdf(name, time, timestampflag='exact', averageyearflag=False):
    """ Read maps from netCDF stacks (forcings, fractions, water demand)

    Read maps from netCDF stacks (forcings, fractions, water demand).
    Maps are read by date, so stacks can start at every date also different from CalendarDayStart.
    Units for stacks can be different from model timestep.
    It can read sub-daily steps.
    timestampflag indicates whether to load data with the exact time stamp ('exact'), or the data with the closest time
    stamp when the exact one is not available ('closest').
    averageyearflag indicates whether to load data from a netcdf file containing one single "average" year (it's used for
    water demand and landuse changes in time).

    :param name: string containing path and name of netCDF file to be read
    :param time: current simulation timestep of the model as integer number (referred to CalendarStartDay)
    :param timestampflag: look for exact time stamp in netcdf file ('exact') or for the closest (left) time stamp available ('closest')
    :param averageyearflag: if True, use "average year" netcdf file over the entire model simulation period
    :returns: content of netCDF map for timestep "time" (mapC)
    :except: if current simulation timestep is not stored in the stack, it stops with error message (if timestampflag='exact')
    """

    filename = name + ".nc" if not name.endswith('.nc') else name
    nf1 = iterOpenNetcdf(filename, "Netcdf map stacks: \n", "r")

    # read information from netCDF file
    variable_name = [k for k in nf1.variables if len(nf1.variables[k].dimensions) == 3][0] # get the variable with 3 dimensions (variable order not relevant)
    t_steps = nf1.variables['time'][:]    # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
    t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
    t_cal = getCalendarType(nf1)
    # CM: get year from time unit in case average year is used
    if averageyearflag:
        # CM: get date of the first step in netCDF file containing average year values
        first_date = num2date(t_steps[0], t_unit, t_cal)
        # CM: get year of the first step in netCDF file containing average year values
        t_ref_year = first_date.year

    begin = Calendar(binding['CalendarDayStart'])
    DtSec = float(binding['DtSec'])
    DtDay = float(DtSec / 86400)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    # get date of current simulation step
    currentDate = Calendar(time)
    if type(currentDate) is not datetime.datetime:
        currentDate = begin + datetime.timedelta(days=(currentDate - 1) * DtDay)

    # if reading from an average year NetCDF stack, ignore the year in current simulation date and change it to the netCDF time unit year
    if averageyearflag:
        try:
            currentDate = currentDate.replace(year=t_ref_year)
        except:
            # CM: if simulation year is leap and average year is not, switch 29/2 with 28/2
            currentDate = currentDate.replace(day=28)
            currentDate = currentDate.replace(year=t_ref_year)

    # get timestep in netCDF file corresponding to current simulation date
    current_ncdf_step = date2num(currentDate, units=t_unit, calendar=t_cal)

    # read netCDF map
    if not (current_ncdf_step in t_steps):
        if (timestampflag == 'exact'):
            # look for exact time stamp when loading data
            msg = "Date " + str(currentDate) + " not stored in " + filename
            raise LisfloodError(msg)
        elif (timestampflag == 'closest'):
            # CM: get the closest value
            current_ncdf_step_new = takeClosest(t_steps, current_ncdf_step)
            # CM: set current_ncdf_step to the closest available time step in netCDF file
            current_ncdf_step = current_ncdf_step_new

    # get index of timestep in netCDF file corresponding to current simulation date
    current_ncdf_index = np.where(t_steps == current_ncdf_step)[0][0]
    mapnp = nf1.variables[variable_name][current_ncdf_index, cutmap[2]:cutmap[3], cutmap[0]:cutmap[1]]
    nf1.close()

    mapC = compressArray(mapnp, pcr=False, name=filename)
    if Flags['check']:
        timename = os.path.basename(name) + str(time)
        checkmap(timename, filename, decompress(mapC), True, 1)
    return mapC


def readnetcdfsparse(name, time, oldmap):
    """
    NO LONGER USED
    load stack of maps 1 at each timestamp in Netcdf format
    """
    try:
        mapC = readnetcdf(name, time)
        find = 1
        # print name+str(time)+"   "+str(find)
    except:
        find = 2
        if oldmap is None:
            for i in range(time - 1, 0, -1):
                try:
                    mapC = readnetcdf(name, i)
                    find = 1
                    break
                except:
                    pass
                # print name+"   "+str(time)+"   "+str(find)+"   "+str(i)
            if find == 2:
                msg = "no map in stack has a smaller time stamp than: " + str(time)
                raise LisfloodError(msg)
        else:
            mapC = oldmap
            if Flags['loud']:
                s = " last_" + (os.path.basename(name)) + str(time)
                # print s,
    return mapC


def checknetcdf(name, start, end):
    """ Check available time steps in netCDF input file
    
    Check available timesteps in netCDF file. Get first and last available timestep in netCDF file and compare with
    first and last computation timestep of the model.
    It can use sub-daily steps.
    
    :param name: string containing path and name of netCDF file
    :param start: initial date or step number of model simulation
    :param end: final date or step of model simulation  
    :return: none
    :raises Exception: stop if netCDF maps do not cover simulation time period
    """

    filename = name + ".nc"
    nf1 = iterOpenNetcdf(filename, "Netcdf map stacks: \n", "r")

    # read information from netCDF file
    t_steps = nf1.variables['time'][:]    # get values for timesteps ([  0.,  24.,  48.,  72.,  96.])
    t_unit = nf1.variables['time'].units  # get unit (u'hours since 2015-01-01 06:00:00')
    t_cal = getCalendarType(nf1)
    if t_cal != binding['calendar_type']:
        print(CalendarInconsistencyWarning(filename, t_cal, binding['calendar_type']))

    # get date of first available timestep in netcdf file
    date_first_step_in_ncdf = num2date(t_steps[0], units=t_unit, calendar=t_cal)
    # get date of last available timestep in netcdf file
    date_last_step_in_ncdf = num2date(t_steps[-1], units=t_unit, calendar=t_cal)

    nf1.close()

    #calendar date start (CalendarDayStart)
    begin = Calendar(binding['CalendarDayStart'])
    DtSec = float(binding['DtSec'])
    DtDay = float(DtSec / 86400)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)


    date_first_sim_step = Calendar(start)
    if type(date_first_sim_step) is not datetime.datetime:
        date_first_sim_step = begin + datetime.timedelta(days=(date_first_sim_step - 1) * DtDay)
    if (date_first_sim_step < date_first_step_in_ncdf):
        msg = "First simulation time step is before first time step in netCDF input data file \n" \
              "File name: "+ filename +"\n" \
              "netCDF start date: "+ date_first_step_in_ncdf.strftime('%d/%m/%Y %H:%M') +"\n" \
              "simulation start date: "+ date_first_sim_step.strftime('%d/%m/%Y %H:%M')
        raise LisfloodError(msg)
        quit(1)

    date_last_sim_step = Calendar(end)
    if type(date_last_sim_step) is not datetime.datetime:
        date_last_sim_step = begin + datetime.timedelta(days=(date_last_sim_step - 1) * DtDay)
    if (date_last_sim_step > date_last_step_in_ncdf):
        msg = "Last simulation time step is after last time step in netCDF input data file \n" \
              "File name: " + filename +"\n" \
              "netCDF last date: " + date_last_step_in_ncdf.strftime('%d/%m/%Y %H:%M') +"\n" \
              "simulation last date: " + date_last_sim_step.strftime('%d/%m/%Y %H:%M')
        raise LisfloodError(msg)
        quit(1)

    return



def generateName(name, time):
    """Returns a filename based on the name and time step passed in.
    The resulting name obeys the 8.3 DOS style format. The time step
    will be added to the end of the filename and be prepended by 0's if
    needed.
    The time step normally ranges from [1, nrTimeSteps].
    The length of the name should be max 8 characters to leave room for
    the time step.
    The name passed in may contain a directory name.
    See also: generateNameS(), generateNameST()
    """
    head, tail = os.path.split(name)
    if re.search("\.", tail):
        msg = "File extension given in '" + name + "' not allowed"
        raise LisfloodError(msg)
    if len(tail) == 0:
        msg = "No filename specified"
        raise LisfloodError(msg)
    if len(tail) > 8:
        msg = "Filename '" + name + "' must be shorter than 8 characters"
        raise LisfloodError(msg)
    if time < 0:
        msg = "Timestep must be larger than 0"
        raise LisfloodError(msg)

    nr = "%d" % (time)
    space = 11 - (len(tail) + len(nr))
    assert space >= 0
    result = "%s%s%s" % (tail, space * "0", nr)
    result = "%s.%s" % (result[:8], result[8:])
    assert len(result) == 12
    return os.path.join(head, result)


def writenet(flag, inputmap, netfile, DtDay,
             value_standard_name, value_long_name, value_unit, data_format,
             startdate, repstepstart, repstepend, frequency=None):

    """ Write a netcdf stack

    :param flag: 0 netCDF file format; ?
    :param inputmap: values to be written to NetCDF file
    :param netfile: name of output file in NetCDF format
    :param DtDay: model timestep (self.var.DtDay)
    :param value_standard_name: variable name to be put into netCDF file
    :param value_long_name: variable long name to be put into netCDF file
    :param value_unit: variable unit to be put into netCDF file
    :param data_format: data format
    :param startdate: reference date to be used to get start date and end date for netCDF file from start step and end step
    :param: repstepstart: first reporting step
    :param: repstepend: final reporting step
    :param frequency:[None,'all','monthly','annual'] save to netCDF stack; None save to netCDF single
    :return: 
    """

    # prefix = netfile.split('/')[-1].split('\\')[-1].split('.')[0]
    prefix = os.path.basename(netfile)
    netfile += ".nc"
    row = np.abs(cutmap[3] - cutmap[2])
    col = np.abs(cutmap[1] - cutmap[0])
    if flag == 0:
        nf1 = iterOpenNetcdf(netfile, "", 'w', format='NETCDF4')
        # general Attributes
        nf1.settingsfile = os.path.realpath(sys.argv[1])
        nf1.date_created = xtime.ctime(xtime.time())
        nf1.Source_Software = 'Lisflood Python'
        nf1.institution = "European Commission DG Joint Research Centre (JRC) - E1, D2 Units"
        nf1.creator_name = "Peter Burek, A de Roo, Johan van der Knijff"
        nf1.source = 'Lisflood output maps'
        nf1.keywords = "Lisflood, EFAS, GLOFAS"
        nf1.Conventions = 'CF-1.6'
        # Dimension
        if 'x' in metadataNCDF.keys():
            lon = nf1.createDimension('x', col)  # x 1000
            longitude = nf1.createVariable('x', 'f8', ('x',))
            for i in metadataNCDF['x']:
                exec('%s="%s"') % ("longitude." + i, metadataNCDF['x'][i])

        if 'lon' in metadataNCDF.keys():
            lon = nf1.createDimension('lon', col)
            longitude = nf1.createVariable('lon', 'f8', ('lon',))
            for i in metadataNCDF['lon']:
                exec('%s="%s"') % ("longitude." + i, metadataNCDF['lon'][i])

        if 'y' in metadataNCDF.keys():
            lat = nf1.createDimension('y', row)  # x 950
            latitude = nf1.createVariable('y', 'f8', ('y',))
            for i in metadataNCDF['y']:
                exec('%s="%s"') % ("latitude." + i, metadataNCDF['y'][i])

        if 'lat' in metadataNCDF.keys():
            lat = nf1.createDimension('lat', row)  # x 950
            latitude = nf1.createVariable('lat', 'f8', ('lat',))
            for i in metadataNCDF['lat']:
                exec('%s="%s"') % ("latitude." + i, metadataNCDF['lat'][i])
            # projection

        if 'laea' in metadataNCDF.keys():
            proj = nf1.createVariable('laea', 'i4')
            for i in metadataNCDF['laea']:
                exec('%s="%s"') % ("proj." + i, metadataNCDF['laea'][i])

        if 'lambert_azimuthal_equal_area' in metadataNCDF.keys():
            proj = nf1.createVariable('lambert_azimuthal_equal_area', 'i4')
            for i in metadataNCDF['lambert_azimuthal_equal_area']:
                exec('%s="%s"') % ("proj." + i, metadataNCDF['lambert_azimuthal_equal_area'][i])
        """
        EUROPE
        proj.grid_mapping_name='lambert_azimuthal_equal_area'
        proj.false_easting=4321000.0
        proj.false_northing=3210000.0
        proj.longitude_of_projection_origin = 10.0
        proj.latitude_of_projection_origin = 52.0
        proj.semi_major_axis = 6378137.0
        proj.inverse_flattening = 298.257223563
        proj.proj4_params = "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 +ellps=GRS80 +units=m +no_defs"
        proj.EPSG_code = "EPSG:3035"
        """
        # Fill variables
        cell = round(pcraster.clone().cellSize(),5)
        xl = round((pcraster.clone().west() + cell / 2),5)
        xr = round((xl + col * cell),5)
        yu = round((pcraster.clone().north() - cell / 2),5)
        yd = round((yu - row * cell),5)
        #lats = np.arange(yu, yd, -cell)
        #lons = np.arange(xl, xr, cell)
        lats = np.linspace(yu, yd, row, endpoint=False)
        lons = np.linspace(xl, xr, col, endpoint=False)
        latitude[:] = lats
        longitude[:] = lons
        if frequency is not None: # output file with "time" dimension
            #Get initial and final dates for data to be stored in nerCDF file
            first_date, last_date = [startdate + datetime.timedelta(days=(int(k) - 1)*DtDay) for k in
                                     (repstepstart, repstepend)]
            # CM: Create time stamps for each step stored in netCDF file
            time_stamps = [first_date + datetime.timedelta(days=d*DtDay) for d in xrange(repstepend - repstepstart +1)]

            units_time = 'days since %s' % startdate.strftime("%Y-%m-%d %H:%M:%S.0")
            steps = (int(binding["DtSec"]) / 86400.) * np.arange(int(binding["StepStart"]) - 1, int(binding["StepEnd"]))
            if frequency != "all":
                dates = num2date(steps, units_time, binding["calendar_type"])
                next_date_times = np.array([j + datetime.timedelta(seconds=int(binding["DtSec"])) for j in dates])
                if frequency == "monthly":
                    months_end = np.array([dates[j].month != next_date_times[j].month for j in xrange(steps.size)])
                    steps = steps[months_end]
                elif frequency == "annual":
                    years_end = np.array([dates[j].year != next_date_times[j].year for j in xrange(steps.size)])
                    steps = steps[years_end]
            nf1.createDimension('time', steps.size)
            time = nf1.createVariable('time', float, ('time'))
            time.standard_name = 'time'
            # time.units ='days since 1990-01-01 00:00:00.0'
            # time.units = 'hours since %s' % startdate.strftime("%Y-%m-%d %H:%M:%S.0")
            # CM: select the time unit according to model time step
            DtDay_in_sec = DtDay * 86400
            if DtDay_in_sec >= 86400:
                # Daily model time steps or larger
                time.units = 'days since %s' % startdate.strftime("%Y-%m-%d %H:%M:%S.0")
            elif DtDay_in_sec >= 3600 and DtDay_in_sec < 86400:
                # CM: hours to days model time steps
                time.units = 'hours since %s' % startdate.strftime("%Y-%m-%d %H:%M:%S.0")
            elif DtDay_in_sec >= 60 and DtDay_in_sec <3600:
                # CM: minutes to hours model time step
                time.units = 'minutes since %s' % startdate.strftime("%Y-%m-%d %H:%M:%S.0")

            time.calendar = binding["calendar_type"]
            nf1.variables["time"][:] = date2num(time_stamps, time.units, time.calendar)
            # for i in metadataNCDF['time']: exec('%s="%s"') % ("time."+i, metadataNCDF['time'][i])
            if 'x' in metadataNCDF.keys():
                value = nf1.createVariable(prefix, data_format, ('time', 'y', 'x'), zlib=True, fill_value=-9999, chunksizes=(1, row, col))
            if 'lon' in metadataNCDF.keys():
                value = nf1.createVariable(prefix, data_format, ('time', 'lat', 'lon'), zlib=True, fill_value=-9999, chunksizes=(1, row, col))
        else:
            if 'x' in metadataNCDF.keys():
                value = nf1.createVariable(prefix, data_format, ('y', 'x'), zlib=True, fill_value=-9999)
            if 'lon' in metadataNCDF.keys():
                # for world lat/lon coordinates
                value = nf1.createVariable(prefix, data_format, ('lat', 'lon'), zlib=True, fill_value=-9999)

        value.standard_name = value_standard_name
        value.long_name = value_long_name
        value.units = value_unit
        for var in metadataNCDF.keys():
            if "esri_pe_string" in metadataNCDF[var].keys():
                value.esri_pe_string = metadataNCDF[var]['esri_pe_string']
    else:
        nf1 = iterOpenNetcdf(netfile, "", 'a',format='NETCDF4')

    mapnp = maskinfo['maskall'].copy()
    mapnp[~maskinfo['maskflat']] = inputmap[:]
    #mapnp = mapnp.reshape(maskinfo['shape']).data
    mapnp = mapnp.reshape(maskinfo['shape'])
    if frequency is not None:
        nf1.variables[prefix][flag, :, :] = mapnp
        #value[flag,:,:]= mapnp
    else:
        # without timeflag
        nf1.variables[prefix][:, :] = mapnp
    nf1.close()

def dumpObject(name, var, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  file_object1 = open(path1, 'w')
  pickle.dump(var, file_object1)
  file_object1.close()

def loadObject(name, num):
  path1 = os.path.join(str(num), 'stateVar', name)
  filehandler1 = open(path1, 'r')
  #read a string from the open file object file and interpret it as a pickle data stream, rec
  var = pickle.load(filehandler1)
  filehandler1.close()
  return(var)

def dumpPCRaster(name, var, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  report(var, path1)

def loadPCRaster(name, num):
  path1 = os.path.join(str(num), 'stateVar',name)
  var = iterReadPCRasterMap(path1)
  return(var)

def perturbState(var, method = "normal", minVal=0, maxVal=1, mu=0, sigma=1, spatial=True, single=True):
  try:
    numVals = len(var)
  except:
    numVals = 1
  if method == "normal":
    if spatial:
      domain = len(var[0])
      out = var
      for i in range(numVals):
        out[i] = np.minimum(np.maximum(np.random.normal(mu, sigma, domain), minVal), maxVal)
    else:
      if single:
        out = np.minimum(np.maximum(np.random.normal(mu, sigma, numVals), minVal), maxVal)
      else:
        out = list(np.minimum(np.maximum(np.random.normal(mu, sigma, numVals), minVal), maxVal))
  if method == "uniform":
    if spatial:
      domain = len(var[0])
      out = var
      for i in range(numVals):
        out[i] = np.random.uniform(minVal, maxVal, domain)
    else:
      if single:
        out = np.random.uniform(minVal, maxVal, numVals)
      else:
        out = list(np.random.uniform(minVal, maxVal, numVals))
  return(out)


def read_tss_header(tssfilename):
    """ Read header of a tss file (used in inflow)
        :param tssfilename  path and name of the tss
        :returns outlets_id  list of column names in tss file 
    """
    with open(tssfilename) as fp:
        rec = fp.readline()
        if rec.split()[0] == 'timeseries':
            # LISFLOOD tss file with header
            # get total number of outlets
            outlets_tot_number = int(fp.readline())
            fp.readline()
            outlets_id = []
            for i in range(0, outlets_tot_number - 1):
                rec = fp.readline()
                rec = int(rec.strip())
                outlets_id.append(rec)  #Lisflood ID code for output points
            # read tss data
            # tssdata = pd.read_table(tssfilename, delim_whitespace=True, header=None, names=outlets_id, index_col=0,
            #                        skiprows=outlets_tot_number + 2)

        else:
            # LISFLOOD tss file without header (table)
            numserie = len(rec.split())
            outlets_id = []
            for i in range(1, numserie):
                outlets_id.append(i)  #Lisflood progressive ID code for output points
            # read tss data
            # tssdata = pd.read_table(tssfilename, delim_whitespace=True, header=None, names=outlets_id, index_col=0)
    fp.close()
    return outlets_id



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *
import netCDF4
import xarray as xr
from pyproj import Proj

def coordinatesLand(eastings_forcing, northings_forcing):
    """"""
    half_cell = maskmapAttr['cell'] / 2.
    top_row = np.where(northings_forcing == maskmapAttr['y'] - half_cell)[0][0]
    left_col = np.where(eastings_forcing == maskmapAttr['x'] + half_cell)[0][0]
    row_slice = slice(top_row, top_row + maskmapAttr['row'])
    col_slice = slice(left_col, left_col + maskmapAttr['col'])
    return [co[row_slice,col_slice][~maskinfo["mask"]] for co in np.meshgrid(eastings_forcing, northings_forcing)]

class miscInitial(object):

    """
    Miscellaneous repeatedly used expressions
    """

    def __init__(self, misc_variable):
        self.var = misc_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the misc module
        """

        if option['gridSizeUserDefined']:

            # <lfoption name="gridSizeUserDefined" choice="1" default="0">
            # If option gridsizeUserDefined is activated, users can specify grid size properties
            # in separate maps. This is useful whenever this information cannot be derived from
            # the location attributes of the base maps (e.g. lat/lon systems or non-equal-area
            # projections)
            # Limitation: always assumes square grid cells (not rectangles!). Size of grid cells
            # may vary across map though

            self.var.PixelLengthPcr = loadmap('PixelLengthUser',pcr=True)
            self.var.PixelLength = compressArray(self.var.PixelLengthPcr)
            # Length of pixel [m]
            # Area of pixel [m2]
            self.var.PixelAreaPcr = loadmap('PixelAreaUser',pcr=True)
            self.var.PixelArea = compressArray(self.var.PixelAreaPcr)

        else:
            # Default behaviour: grid size is derived from location attributes of
            # base maps. Requirements:
            # - Maps are in some equal-area projection
            # - Length units meters
            # - All grid cells have the same size

            # Length of pixel [m]
            #self.var.PixelLength = celllength()
            self.var.PixelLengthPcr = celllength()
            self.var.PixelLength = maskmapAttr['cell']

            # Area of pixel [m2]
            self.var.PixelAreaPcr = self.var.PixelLength ** 2
            self.var.PixelArea=np.empty(maskinfo['mapC'])
            self.var.PixelArea.fill(self.var.PixelLength ** 2)

#            self.var.PixelArea = spatial(self.var.PixelArea)
            # Convert to spatial expresion (otherwise this variable cannnot be
            # used in areatotal function)

# -----------------------------------------------------------------
        # Miscellaneous repeatedly used expressions (as suggested by GF)

        self.var.InvPixelLength = 1.0 / self.var.PixelLength
        # Inverse of pixel size [1/m]
        self.var.DtSec = loadmap('DtSec')
        self.var.DtDay = self.var.DtSec / 86400
        # Time step, expressed as fraction of day (used to convert
        # rate variables that are expressed as a quantity per day to
        # into an amount per time step)
        self.var.InvDtSec = 1 / self.var.DtSec
        # Inverse of time step [1/s]
        self.var.InvDtDay = 1 / self.var.DtDay
        # Inverse of time step [1/d]
        self.var.DtSecChannel = loadmap('DtSecChannel')
        # Sub time step used for kinematic wave channel routing [seconds]
        # within the model,the smallest out of DtSecChannel and DtSec is used

        self.var.MMtoM = 0.001
        # Multiplier to convert wate depths in mm to meters
        self.var.MtoMM = 1000
        # Multiplier to convert wate depths in meters to mm
        self.var.MMtoM3 = 0.001 * self.var.PixelArea
        # self.var.MMtoM3=0.001*float(celllength())**2
        # Multiplier to convert water depths in mm to cubic
        # metres
        self.var.M3toMM = 1 / self.var.MMtoM3
        # Multiplier to convert from cubic metres to mm water slice

        self.var.GwLoss = loadmap('GwLoss')
        self.var.GwPerc = np.maximum(loadmap('GwPercValue'), self.var.GwLoss)
        # new Gwloss  PB 12.11.2009
        # if GWloss > GwPercValue -> GwPerc = GwLoss
        self.var.GwPercStep = self.var.GwPerc * self.var.DtDay
        # Percolation from upper to lower groundwater zone, expressed as
        # amount per time step
        self.var.GwLossStep = self.var.GwLoss * self.var.DtDay
        # change similar to GwPercStep

        # ************************************************************
        # ***** Some additional stuff
        # ************************************************************
        # date of the first possible model run
        # computation of model steps is referred to CalendarStartDay
        self.var.CalendarDayStart = Calendar(binding['CalendarDayStart'])
        try:
           # number of time step or date of the state map to be used to initialize model run
           timestepInit.append(binding["timestepInit"])
        except: pass

        self.var.PrScaling = loadmap('PrScaling')
        self.var.CalEvaporation = loadmap('CalEvaporation')

        self.var.Precipitation = None
        self.var.Tavg = None
        self.var.ETRef = None
        self.var.ESRef = None
        self.var.EWRef = None
        # setting meteo data to none - is this necessary?

        self.var.DayCounter= 0.0
        self.var.MonthETpot= globals.inZero
        self.var.MonthETact= globals.inZero
        self.var.MonthWDemand= globals.inZero
        self.var.MonthWUse= globals.inZero
        self.var.MonthWDemand= globals.inZero
        self.var.MonthDis= globals.inZero
        self.var.MonthInternalFlow =  globals.inZero

        self.var.TotalInternalFlowM3 = globals.inZero
        self.var.PerMonthInternalFlowM3 = globals.inZero
        # total freshwater generated in the sub-area (m3), basically local P-ET-Storage
        self.var.TotalExternalInflowM3 = globals.inZero
        self.var.PerMonthExternalInflowM3 = globals.inZero
        # Total channel inflow (m3) from inlet points
        self.var.PerMonthWaterDemandM3 = globals.inZero
        self.var.PerMonthWaterUseM3 = globals.inZero

        self.var.FlagDemandBiggerUse = scalar(0.0)

        self.var.TotWEI = scalar(0.0)
        self.var.TotlWEI = scalar(0.0)
        self.var.TotCount = scalar(0.0)

        self.var.SumETpot = globals.inZero
        self.var.SumETpotact = globals.inZero

        # Read the latitude (radians) from the precipitation forcing NetCDF file
        with xr.open_dataset(binding["PrecipitationMaps"] + ".nc") as nc:
            if all([co in nc.dims for co in ("x", "y")]):
                try:
                    proj_var = [v for v in nc.data_vars.keys() if 'proj4_params' in nc[v].attrs.keys()][0]  # look for the projection variable
                except IndexError:
                    raise Exception("If using projected coordinates (x, y), a variable with the 'proj4_params' attribute must be included in the precipitation file!")
                projection = Proj(nc[proj_var].attrs['proj4_params']) # projection object obtained from the PROJ4 string
                _, lat_deg = projection(*coordinatesLand(nc.x.values, nc.y.values), inverse=True) # latitude (degrees)
            else:
                _, lat_deg = coordinatesLand(nc.lon.values, nc.lat.values)  # latitude (degrees)
        self.var.lat_rad = np.radians(lat_deg)  # latitude (radians)



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *

class groundwater(object):

    """
    # ************************************************************
    # ***** GROUNDWATER   *****************************************
    # ************************************************************
    """

    def __init__(self, groundwater_variable):
        self.var = groundwater_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the groundwater module
        """

        # ********************************************
        # ****  Groundwater
        # ********************************************
        UpperZoneTimeConstant = loadmap('UpperZoneTimeConstant')
        LowerZoneTimeConstant = loadmap('LowerZoneTimeConstant')
        self.var.UpperZoneK = np.minimum(self.var.DtDay * (1 / UpperZoneTimeConstant), 1)
        self.var.LowerZoneK = np.minimum(self.var.DtDay * (1 / LowerZoneTimeConstant), 1)
        # Reservoir constants for upper- and lower zone
        # Minimize statement in case time constants greater than DtDay (again
        # unlikely, but you never know..)

        if option['InitLisflood']:
            LZAvInflowGuess = self.var.GwPerc - self.var.GwLoss
        else:
            try:
                LZAvInflowGuess = np.minimum(loadmap('LZAvInflowMap'), self.var.GwPerc - self.var.GwLoss)
            except:
                msg = "Repeat InitLisflood: LZAvin.map does not exists or is not compatible with mask map"
                raise LisfloodError(msg)
            # Estimated average inflow rate into lower zone [mm/day]
            # By definition value cannot exceed GwPercValue!
            # LZAvinflowEstimate can be based on climatic data or baseflow
            # analysis, or directly from model (pre)run using "repLZAvInflowMap"
            # option.
        LZAvInflowGuess = makenumpy(LZAvInflowGuess)

        LZSteady = LZAvInflowGuess * LowerZoneTimeConstant
        # Steady-state amount of water in lower store [mm]
        LZInitValue = loadmap('LZInitValue')

        self.var.LZ = np.where(LZInitValue == -9999, LZSteady, LZInitValue)
        # Initialise lower store with steady-state value
        # if LZInitValue is set to -9999
        
        # Water in lower store [mm]
        self.var.LZThreshold = loadmap('LZThreshold')
        # lz threshold =if lz falls below this there is no outflow to the channel from lz

        # UZInitValue=scalar(loadmap('UZInitValue'))
        # UZInit=ifthen(defined(self.var.MaskMap),UZInitValue)
        self.var.UZ = defsoil('UZInitValue', 'UZForestInitValue', 'UZIrrigationInitValue')
        # Water in upper store [mm]

        # TotalGroundWaterInit=ifthenelse(defined(self.var.MaskMap),self.var.UZ+self.var.LZ,scalar(0.0))
        # total amount of initial groundwater (or, more correctly, water stored
        # in Upper and Lower Zone [mm]

        # Initialising cumulative output variables
        # These are all needed to compute the cumulative mass balance error

        self.var.GwLossCUM = globals.inZero.copy()
        # Cumulative groundwater loss [mm]
        self.var.LZInflowCUM = globals.inZero.copy()
        # Cumulative lower zone inflow [mm]
        # Needed for calculation of average LZ inflow (initialisation)

        self.var.GwPercUZLZ = defsoil(globals.inZero.copy())
        self.var.GwLossLZ = globals.inZero.copy()
        self.var.UZOutflow = defsoil(globals.inZero.copy())
        self.var.LZOutflow = globals.inZero.copy()


    def dynamic(self):
        # outflow from LZ to channel stops when LZ is below its threshold. LZ can be below its threshold because of water abstractions
        self.var.LZOutflow = np.minimum(self.var.LowerZoneK * self.var.LZ,self.var.LZ - self.var.LZThreshold)
        # Outflow out of lower zone [mm per model timestep]

        self.var.LZOutflow = np.maximum(self.var.LZOutflow, 0)        
        # No outflow if LZ is below threshold
        self.var.LZOutflowToChannel = self.var.LZOutflow
        # And the remaining amount goes to the channel

        self.var.LZ -= self.var.LZOutflow
        # Update upper-, lower zone storage

        self.var.UZOutflowPixel = self.var.deffraction(self.var.UZOutflow)
          # outflow from upper zone as pixel flow
          # to surface routing

        self.var.GwPercUZLZPixel = self.var.deffraction(self.var.GwPercUZLZ)
          #  Compute pixel-average flux
          # to Lower Zone

        self.var.LZ += self.var.GwPercUZLZPixel
      	# (ground)water in lower response box [mm]

        self.var.GwLossLZ = np.maximum(np.minimum(self.var.GwLossStep,self.var.LZ),0.0)
      	# same method as GWPerc
      	# maximum value is controlled by GwLossStep (which is GWLoss*DtDay)
		# prevention to go negative, when LZ is negative
        self.var.LZ = self.var.LZ-self.var.GwLossLZ
      	# (ground)water in lower response box [mm]

        self.var.LZInflowCUM += (self.var.GwPercUZLZPixel-self.var.GwLossLZ)
     	# cumulative inflow into lower zone (can be used to improve
        self.var.LZInflowCUM = np.maximum(self.var.LZInflowCUM,0.0)
	    # LZInflowCUM would become negativ, if LZInit is set to high
	    # therefore this line is preventing LZInflowCUM getting negativ

        self.var.GwLossPixel = self.var.GwLossLZ
	    # from GROUNDWATER TRANSFER to here
	    # Compute pixel-average fluxes
        self.var.GwLossCUM += self.var.GwLossPixel
	    # Accumulated groundwater loss over simulation period [mm]
        self.var.LZAvInflow = (self.var.LZInflowCUM * self.var.InvDtDay) / self.var.TimeSinceStart
	    # Average inflow into lower zone over executed time steps [mm/day]

        self.var.LZOutflowToChannelPixel = self.var.LZOutflowToChannel



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *

class structures(object):
    """
    ************************************************************
    ***** STRUCTURES
    ************************************************************
    Structures such as reservoirs and lakes are modelled by interrupting the channel flow paths
    for the kinematic wave (LddKinematic). The current block of code produces:
    1. A modified LddKinematic, where a pit (sink) is inserted at all cells immediately upstream of
       a structure (can be more than 1 per structure, in case of junction of multiple branches)
     2. LddStructuresKinematic, equal to the "old" (unmodified) LddKinematic (needed to
        connect the inflow and outflow points of each structure)
    """

    def __init__(self, structures_variable):
        self.var = structures_variable


# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
    def initial(self):
        """ initial part of the structures module
        """
        self.var.LddStructuresKinematic = self.var.LddKinematic

        if not(option['InitLisflood']):
            # not done in Init Lisflood
            IsUpsOfStructureKinematic = downstream(self.var.LddKinematic,
                cover(boolean(decompress(self.var.IsStructureKinematic)), pcraster.boolean(0)))
            # Get all pixels just upstream of kinematic structure locations
            self.var.IsUpsOfStructureKinematicC = compressArray(IsUpsOfStructureKinematic)
            # Unmodified version of LddKinematic is needed to connect inflow and outflow points
            # of each structure (called LddStructuresKinematic now)
            self.var.LddKinematic = lddrepair(ifthenelse(IsUpsOfStructureKinematic, 5
                                    , self.var.LddKinematic))
            # Cells just upstream of each structure are treated as pits in the kinematic wave
            # channel routing



 #           lddC = compressArray(self.var.LddKinematic)
 #           inAr = decompress(np.arange(maskinfo['mapC'][0],dtype="int32"))
            # giving a number to each non missing pixel as id
 #           self.var.downKin = compressArray(downstream(self.var.LddKinematic,inAr))
            # each upstream pixel gets the id of the downstream pixel
 #           self.var.downKin[lddC==5] = maskinfo['mapC'][0]
            # all pits gets a high number




"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

 ######################################################################

 ##       ####  ######  ######## ##        #######   #######  ########
 ##        ##  ##    ## ##       ##       ##     ## ##     ## ##     ##
 ##        ##  ##       ##       ##       ##     ## ##     ## ##     ##
 ##        ##   ######  ######   ##       ##     ## ##     ## ##     ##
 ##        ##        ## ##       ##       ##     ## ##     ## ##     ##
 ##        ##  ##    ## ##       ##       ##     ## ##     ## ##     ##
 ######## ####  ######  ##       ########  #######   #######  ########

######################################################################
"""

__authors__ = "Ad de Roo, Peter Burek, Johan van der Knijff, Niko Wanders"
__version__ = "Version: 2.02"
__date__ = "23 Apr 2016"
__copyright__ = "Copyright 2016, European Commission - Joint Research Centre"
__maintainer__ = "Ad de Roo"
__status__ = "Operation"


# to work with the new grid engine JRC - workaround with error on pyexpat
from pyexpat import *
import xml.dom.minidom
from netCDF4 import Dataset
from pcraster import *
from pcraster.framework import *

from Lisflood_initial import *
from Lisflood_dynamic import *
from Lisflood_monteCarlo import *
from Lisflood_EnKF import *


class LisfloodModel(LisfloodModel_ini, LisfloodModel_dyn, LisfloodModel_monteCarlo, LisfloodModel_EnKF):
    """ Joining the initial and the dynamic part
        of the Lisflood model
    """

# ==================================================
# ============== LISFLOOD execute ==================
# ==================================================

def Lisfloodexe():

    optionBinding(settings, optionxml)
    # read all the possible option for modelling and for generating output
    # read the settingsfile with all information about the catchments(s)
    # and the choosen option for mdelling and output
    bindkey = sorted(binding.keys())

    #os.chdir(outputDir[0])
    # this prevent from using relative path in settings!


    checkifDate('StepStart','StepEnd')

    if option['InitLisflood']: print "INITIALISATION RUN"
    print "Start - End: ",modelSteps[0]," - ", modelSteps[1]
    if Flags['loud']:
        print"%-6s %10s %11s\n" %("Step","Date","Discharge"),

    Lisflood = LisfloodModel()
    stLisflood = DynamicFramework(Lisflood, firstTimestep=modelSteps[0], lastTimeStep=modelSteps[1])
    stLisflood.rquiet = True
    stLisflood.rtrace = False


    """
    ----------------------------------------------
    Monte Carlo and Ensemble Kalman Filter setting
    ----------------------------------------------
    """

    try:
        EnKFset = option['EnKF']
    except:
        EnKFset = 0
    try:
        MCset = option['MonteCarlo']
    except:
        MCset = 0
    if option['InitLisflood']:
        MCset = 0
        EnKFset = 0
    if EnKFset and not MCset:
        msg = "Trying to run EnKF with only 1 ensemble member \n"
        raise LisfloodError(msg)
    if EnKFset and FilterSteps[0] == 0:
        msg = "Trying to run EnKF without filter timestep specified \nRunning LISFLOOD in Monte Carlo mode \n"
        print LisfloodWarning(msg)
        EnKFset = 0
    if MCset and EnsMembers[0] <= 1:
        msg = "Trying to run Monte Carlo simulation with only 1 member \nRunning LISFLOOD in deterministic mode \n"
        print LisfloodWarning(msg)
        MCset = 0
    if MCset:
        mcLisflood = MonteCarloFramework(stLisflood, nrSamples=EnsMembers[0])
        if nrCores[0] > 1:
            mcLisflood.setForkSamples(True, nrCPUs=nrCores[0])
        if EnKFset:
            kfLisflood = EnsKalmanFilterFramework(mcLisflood)
            kfLisflood.setFilterTimesteps(FilterSteps)
            print LisfloodRunInfo(mode = "Ensemble Kalman Filter", outputDir = outputDir[0], Steps = len(FilterSteps), ensMembers=EnsMembers[0], Cores=nrCores[0])
            kfLisflood.run()
        else:
            print LisfloodRunInfo(mode = "Monte Carlo", outputDir = outputDir[0], ensMembers=EnsMembers[0], Cores=nrCores[0])
            mcLisflood.run()
    else:
        """
        ----------------------------------------------
        Deterministic run
        ----------------------------------------------
        """
        print LisfloodRunInfo(mode = "Deterministic", outputDir = outputDir[0])
        stLisflood.run()
    # cProfile.run('stLisflood.run()')
    # python -m cProfile -o  l1.pstats lisf1.py settingsNew3.xml
    # gprof2dot -f pstats l1.pstats | dot -Tpng -o callgraph.png

    if Flags['printtime']:
        print "\n\nTime profiling"
        print "%2s %-17s %10s %8s" %("No","Name","time[s]","%")
        div = 1
        timeSum = np.array(timeMesSum)
        if MCset:
            div = div * EnsMembers[0]
        if EnKFset:
            div = div * (len(FilterSteps)+1)
        if EnKFset or MCset:
            timePrint = np.zeros(len(timeSum)/div)
            for i in range(len(timePrint)):
                timePrint[i] = np.sum(timeSum[range(i, len(timeSum), len(timeSum)/div)])
        else:
            timePrint = timeSum
        for i in xrange(len(timePrint)):
            print "%2i %-17s %10.2f %8.1f"  %(i,timeMesString[i],timePrint[i],100 * timePrint[i] / timePrint[-1])
    i=1

# ==================================================
# ============== USAGE ==============================
# ==================================================


def usage():
    """ prints some lines describing how to use this program
        which arguments and parameters it accepts, etc
    """
    print 'LisfloodPy - Lisflood using pcraster Python framework'
    print 'Authors: ', __authors__
    print 'Version: ', __version__
    print 'Date: ', __date__
    print 'Status: ', __status__
    print """
    Arguments list:
    settings.xml     settings file

    -q --quiet       output progression given as .
    -v --veryquiet   no output progression is given
    -l --loud        output progression given as time step, date and discharge
    -c --check       input maps and stack maps are checked, output for each input map BUT no model run
    -h --noheader    .tss file have no header and start immediately with the time series
    -t --printtime   the computation time for hydrological modules are printed
    """
    sys.exit(1)

def headerinfo():

   print "LisfloodPy ",__version__," ",__date__,
   print """
Water balance and flood simulation model for large catchments\n
(C) Institute for Environment and Sustainability
    Joint Research Centre of the European Commission
    TP122, I-21020 Ispra (Va), Italy\n"""


# ==================================================
# ============== MAIN ==============================
# ==================================================

if __name__ == "__main__":

    if len(sys.argv) < 2:
        usage()


    LF_Path = os.path.dirname(sys.argv[0])
    LF_Path = os.path.abspath(LF_Path)
    optionxml = os.path.normpath(LF_Path + "/OptionTserieMaps.xml")

    settings = sys.argv[1]    # setting.xml file

    args = sys.argv[2:]
    globalFlags(args)
    # setting of global flag e.g checking input maps, producing more output information
    if not(Flags['veryquiet']) and not(Flags['quiet']) : headerinfo()
    Lisfloodexe()



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from pcraster import *
from pcraster.framework import *

from lisflood.global_modules.add1 import *
from lisflood.global_modules.globals import *


class waterlevel(object):
    """
    # ************************************************************
    # ***** WATER LEVEL    *****************************************
    # ************************************************************
    """

    def __init__(self, waterlevel_variable):
        self.var = waterlevel_variable

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the water level module
        """

        if option['simulateWaterLevels']:
            self.var.FloodPlainWidth = loadmap('FloodPlainWidth')

    def dynamic(self):
        """ dynamic part of the water level module
        """

        # Additional computations for channel and floodplain geometry
        # Activating this option doesn't affect LISFLOOD's behaviour in any way, option only
        # included to allow reporting of water level maps / time series
        # Actual reporting activated using separate options!
        if option['simulateWaterLevels']:
            ChanCrossSectionArea = np.where(
                self.var.IsChannelKinematic,
                np.minimum(self.var.TotalCrossSectionArea, self.var.TotalCrossSectionAreaBankFull),
                0
            )
            # Cross-sectional area for channel only (excluding floodplain) [sq m]
            FloodPlainCrossSectionArea = self.var.TotalCrossSectionArea - ChanCrossSectionArea
            # Floodplain area [sq m]
            ChanWaterDepth = 2 * ChanCrossSectionArea / (self.var.ChanUpperWidth + self.var.ChanBottomWidth)
            # Water level in channel [m]
            FloodPlainWaterDepth = FloodPlainCrossSectionArea / self.var.FloodPlainWidth
            # Water level on floodplain [m]
            WaterLevelKin = ChanWaterDepth + FloodPlainWaterDepth
            # Total water level [m]
            self.var.WaterLevel = np.where(self.var.IsChannelKinematic, WaterLevelKin, 0)
            # Use WaterLevelKin if kinematic wave routing is used ...
            if option['dynamicWave']:
                self.var.WaterLevel = self.var.WaterLevelDyn



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.
"""

__authors__ = "Ad de Roo, Emiliano Gelati, Peter Burek, Johan van der Knijff, Niko Wanders"
__version__ = "2.8.15"
__date__ = "21 May 2019"
__copyright__ = "Copyright 2019, European Commission - Joint Research Centre"
__maintainer__ = "Emiliano Gelati, Domenico Nappo, Valerio Lorini, Lorenzo Mentaschi, Ad de Roo"
__status__ = "Operation"



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *
import gc


class LisfloodModel_dyn(DynamicModel):

    # =========== DYNAMIC ====================================================

    def dynamic(self):
        """ Dynamic part of LISFLOOD
            calls the dynamic part of the hydrological modules
        """
        del timeMes[:]
        # get time for operation "Start dynamic"
        timemeasure("Start dynamic")
        # date corresponding to the model time step (yyyy-mm-dd hh:mm:ss)
        self.CalendarDate = self.CalendarDayStart + datetime.timedelta(days=(self.currentTimeStep()-1) * self.DtDay)
        # day of the year corresponding to the model time step
        self.CalendarDay = int(self.CalendarDate.strftime("%j"))
        #correct method to calculate the day of the year

        # model time step
        i = self.currentTimeStep()
        if i==1:    globals.cdfFlag = [0, 0, 0, 0 ,0 ,0,0]
          # flag for netcdf output for all, steps and end
          # set back to 0,0,0,0,0,0 if new Monte Carlo run

        self.TimeSinceStart = self.currentTimeStep() - self.firstTimeStep() + 1

        if Flags['loud']:
            print "%-6i %10s" %(self.currentTimeStep(),self.CalendarDate.strftime("%d/%m/%Y %H:%M")) ,
        else:
            if not(Flags['check']):
                if (Flags['quiet']) and (not(Flags['veryquiet'])):
                    sys.stdout.write(".")
                if (not(Flags['quiet'])) and (not(Flags['veryquiet'])):
                    # Print step number and date to console
                    sys.stdout.write("\r%d" % i), sys.stdout.write("%s" % " - "+self.CalendarDate.strftime("%d/%m/%Y %H:%M"))
                    sys.stdout.flush()

        # ************************************************************
        """ up to here it was fun, now the real stuff starts
        """
        # readmeteo.py
        self.readmeteo_module.dynamic()     
        timemeasure("Read meteo") # 1. timing after read input maps

        if Flags['check']: return  # if check than finish here

        """ Here it starts with hydrological modules:
        """

        # ***** READ land use fraction maps***************************
        self.landusechange_module.dynamic()

        # ***** READ LEAF AREA INDEX DATA ****************************
        self.leafarea_module.dynamic()

        # ***** READ variable water fraction ****************************
        self.evapowater_module.dynamic_init()

        # ***** READ INFLOW HYDROGRAPHS (OPTIONAL)****************
        self.inflow_module.dynamic()
        timemeasure("Read LAI") # 2. timing after LAI and inflow

        # ***** RAIN AND SNOW *****************************************
        self.snow_module.dynamic()
        timemeasure("Snow")  # 3. timing after LAI and inflow

        # ***** FROST INDEX IN SOIL **********************************
        self.frost_module.dynamic()
        timemeasure("Frost")  # 4. timing after frost index

        # ************************************************************
        # ****Looping soil 2 times - second time for forest fraction *
        # ************************************************************

        for soilLoop in xrange(3):
            self.soilloop_module.dynamic(soilLoop)
            # soil module is repeated 2 times:
            # 1. for remaining areas: no forest, no impervious, no water
            # 2. for forested areas
            timemeasure("Soil",loops = soilLoop + 1) # 5/6 timing after soil

        # -------------------------------------------------------------------
        # -------------------------------------------------------------------

        # ***** ACTUAL EVAPORATION FROM OPEN WATER AND SEALED SOIL ***
        self.opensealed_module.dynamic()

        # *********  WATER USE   *************************
        self.riceirrigation_module.dynamic()
        self.waterabstraction_module.dynamic()
        timemeasure("Water abstraction")

        # ***** Calculation per Pixel ********************************
        self.soil_module.dynamic_perpixel()
        timemeasure("Soil done")

        self.groundwater_module.dynamic()
        timemeasure("Groundwater")

        # ************************************************************
        # ***** STOP if no routing is required    ********************
        # ************************************************************
        if option['InitLisfloodwithoutSplit']:
            # InitLisfloodwithoutSplit
            # Very fast InitLisflood
            # it is only to compute Lzavin.map and skip completely the routing component
            self.output_module.dynamic() # only lzavin

            timemeasure("After fast init")
            for i in xrange(len(timeMes)):
                if self.currentTimeStep() == self.firstTimeStep():
                   timeMesSum.append(timeMes[i] - timeMes[0])
                else: timeMesSum[i] += timeMes[i] - timeMes[0]

            return


        # *********  EVAPORATION FROM OPEN WATER *************
        self.evapowater_module.dynamic()
        timemeasure("open water eva.")

        # ***** ROUTING SURFACE RUNOFF TO CHANNEL ********************
        self.surface_routing_module.dynamic()
        timemeasure("Surface routing")  # 7 timing after surface routing

        # ***** POLDER INIT **********************************
        self.polder_module.dynamic_init()

        # ***** INLETS INIT **********************************
        self.inflow_module.dynamic_init()
        timemeasure("Before routing")  # 8 timing before channel routing

        # ************************************************************
        # ***** LOOP ROUTING SUB TIME STEP   *************************
        # ************************************************************
        self.sumDisDay = globals.inZero.copy()
        # sums up discharge of the sub steps
        for NoRoutingExecuted in xrange(self.NoRoutSteps):
            self.routing_module.dynamic(NoRoutingExecuted)
            #   routing sub steps
        timemeasure("Routing",loops = NoRoutingExecuted + 1)  # 9 timing after routing

        # ----------------------------------------------------------------------

        if option['inflow']:
            self.QInM3Old = self.QInM3
            # to calculate the parts of inflow for every routing timestep
            # for the next timestep the old inflow is preserved
            self.sumIn += self.QInDt*self.NoRoutSteps

        # if option['simulatePolders']:
        # ChannelToPolderM3=ChannelToPolderM3Old;

        if option['InitLisflood'] or (not(option['SplitRouting'])):
            # kinematic routing
            self.ChanM3 = self.ChanM3Kin.copy()
                # Total channel storage [cu m], equal to ChanM3Kin
        else:
            # split routing
            self.ChanM3 = self.ChanM3Kin + self.Chan2M3Kin - self.Chan2M3Start #originale
            #self.ChanM3 = self.ChanM3Kin + self.Chan2M3Kin


        # Avoid negative values in ChanM3 and TotalCrossSectionArea
        self.ChanM3 = np.where(self.ChanM3 > 0, self.ChanM3, 0)


            # Total channel storage [cu m], equal to ChanM3Kin
                # sum of both lines
            #CrossSection2Area = pcraster.max(scalar(0.0), (self.Chan2M3Kin - self.Chan2M3Start) / self.ChanLength)

        self.sumDis += self.sumDisDay
        self.ChanQAvg = self.sumDisDay/self.NoRoutSteps
        self.TotalCrossSectionArea = self.ChanM3 * self.InvChanLength
            # Total volume of water in channel per inv channel length
            # New cross section area (kinematic wave)
            # This is the value after the kinematic wave, so we use ChanM3Kin here
            # (NOT ChanQKin, which is average discharge over whole step, we need state at the end of all iterations!)

        timemeasure("After routing")  # 10 timing after channel routing

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if not(option['dynamicWave']):
            # Dummy code if dynamic wave is not used, in which case the total cross-section
            # area equals TotalCrossSectionAreaKin, ChanM3 equals ChanM3Kin and
            # ChanQ equals ChanQKin
            WaterLevelDyn = -9999
            # Set water level dynamic wave to dummy value (needed

        if option['InitLisflood'] or option['repAverageDis']:
            self.CumQ += self.ChanQ
            self.avgdis = self.CumQ/self.TimeSinceStart
            # to calculate average discharge

        self.DischargeM3Out += np.where(self.AtLastPointC ,self.ChanQ * self.DtSec,0)
           # Cumulative outflow out of map

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        # Calculate water level
        self.waterlevel_module.dynamic()

        # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        # ************************************************************
        # *******  Calculate CUMULATIVE MASS BALANCE ERROR  **********
        # ************************************************************
        self.waterbalance_module.dynamic()



        self.indicatorcalc_module.dynamic()



        # ************************************************************
        # ***** WRITING RESULTS: TIME SERIES AND MAPS ****************
        # ************************************************************

        self.output_module.dynamic()
        timemeasure("Water balance")


        # debug 
        # Print value of variables after computation (from state files)
        if Flags['debug']:
            nomefile = 'Debug_out_'+str(self.currentStep)+'.txt'
            ftemp1 = open(nomefile, 'w+')
            nelements = len(self.ChanM3)
            for i in range(0,nelements-1):
                if  hasattr(self,'CrossSection2Area'):
                    print >> ftemp1, i, self.TotalCrossSectionArea[i], self.CrossSection2Area[i], self.ChanM3[i], \
                    self.Chan2M3Kin[i]
                else:
                    print >> ftemp1, i, self.TotalCrossSectionArea[i], self.ChanM3[i]
            ftemp1.close()



        ### Report states if EnKF is used and filter moment
        self.stateVar_module.dynamic()
        timemeasure("State report")

        timemeasure("All dynamic")


        for i in xrange(len(timeMes)):
            if self.currentTimeStep() == self.firstTimeStep():
                timeMesSum.append(timeMes[i] - timeMes[0])
            else: timeMesSum[i] += timeMes[i] - timeMes[0]



        self.indicatorcalc_module.dynamic_setzero()
           # setting monthly and yearly dindicator to zero at the end of the month (year)


        # garbage collector added to free memory at the end of computation step
        gc.collect()





        """
     # self.var.WaterRegionOutflowPoints self.var.WaterRegionInflowPoints
        report(self.WaterRegionOutflowPoints,'D:\Lisflood_runs\LisfloodWorld2\out\wateroutpt.map')
        report(self.WaterRegionInflowPoints,'D:\Lisflood_runs\LisfloodWorld2\out\waterInpt.map')

        # report(self.map2,'mapx.map')
        # self.Tss['UZTS'].sample(Precipitation)
        # self.report(self.Precipitation,binding['TaMaps'])





       #WUse=(WUse*PixelArea*0.001)/2592000;
       # if mm maps are used:
	   # mm per month to m3/s  : x Pixelarea * mmtom / sec in a month

        #self.SumETpot += self.ETRef
        #self.SumET = SumET + MonthETact + TaInterception + TaPixel + ESActPixel + EvaAddM3 * M3toMM;
        #SumTrun += ToChanM3Runoff;



        WUse = decompress((self.TotalAbstractionFromGroundwaterM3 + self.TotalAbstractionFromSurfaceWaterM3) * self.DtSec)
        WaterDemandM3= areatotal(cover(WUse,0.0),wreg)
        WaterUseM3 = areatotal(cover(decompress(self.WUseAddM3),0.0),wreg)
        self.FlagDemandBiggerUse = self.FlagDemandBiggerUse + ifthenelse((WaterDemandM3*0.9) > WaterUseM3,scalar(1.0),scalar(0.0))




        self.TotWEI = self.TotWEI + ifthenelse(EndOfMonth > 0,WEI_Use,scalar(0.0))
        self.TotWEI = ifthenelse(self.currentTimeStep() < 365,scalar(0.0),self.TotWEI)
        self.TotlWEI =  self.TotlWEI + ifthenelse(EndOfMonth > 0,lWEI_Use,scalar(0.0))
        self.TotlWEI = ifthenelse(self.currentTimeStep() < 365,scalar(0.0),self.TotlWEI)
        self.TotCount = self.TotCount + ifthenelse(EndOfMonth > 0 ,scalar(1.0),scalar(0.0))
        self.TotCount = ifthenelse(self.currentTimeStep() < 365,scalar(0.0),self.TotCount)


        # self.CalendarDate.strftime("%d/%m/%Y"))
        """




"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from lisflood.global_modules.add1 import *


class readmeteo(object):

    """
     # ************************************************************
     # ***** READ METEOROLOGICAL DATA              ****************
     # ************************************************************
    """

    def __init__(self, readmeteo_variable):
        self.var = readmeteo_variable

        if option['readNetcdfStack']:
            # checking if time period in netCDF files (forcings) includes simulation period
            checknetcdf(binding['PrecipitationMaps'], binding['StepStart'], binding['StepEnd'] )
            checknetcdf(binding['TavgMaps'], binding['StepStart'], binding['StepEnd'] )
            checknetcdf(binding['ET0Maps'], binding['StepStart'], binding['StepEnd'] )
         #   checknetcdf(binding['ES0Maps'], binding['StepStart'], binding['StepEnd'] )
            checknetcdf(binding['E0Maps'], binding['StepStart'], binding['StepEnd'] )


# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def dynamic(self):
        """ dynamic part of the readmeteo module
            read meteo input maps
        """


        # ************************************************************
        # ***** READ METEOROLOGICAL DATA *****************************
        # ************************************************************
        if option['readNetcdfStack']:
            # Read from NetCDF stack files
            self.var.Precipitation = readnetcdf(binding['PrecipitationMaps'], self.var.currentTimeStep()) * self.var.DtDay * self.var.PrScaling
            self.var.Tavg = readnetcdf(binding['TavgMaps'], self.var.currentTimeStep())
            self.var.ETRef = readnetcdf(binding['ET0Maps'], self.var.currentTimeStep()) * self.var.DtDay * self.var.CalEvaporation
            # self.var.ESRef = readnetcdf(binding['ES0Maps'], self.var.currentTimeStep()) * self.var.DtDay * self.var.CalEvaporation
            self.var.EWRef = readnetcdf(binding['E0Maps'], self.var.currentTimeStep()) * self.var.DtDay * self.var.CalEvaporation
        else:
            # Read from stack of maps in Pcraster format
            self.var.Precipitation = readmapsparse(binding['PrecipitationMaps'], self.var.currentTimeStep(), self.var.Precipitation) * self.var.DtDay * self.var.PrScaling
            # precipitation (conversion to [mm] per time step)
            self.var.Tavg = readmapsparse(binding['TavgMaps'], self.var.currentTimeStep(), self.var.Tavg)
            # average DAILY temperature (even if you are running the model on say an hourly time step) [degrees C]
            self.var.ETRef = readmapsparse(binding['ET0Maps'], self.var.currentTimeStep(), self.var.ETRef) * self.var.DtDay * self.var.CalEvaporation
            # daily reference evapotranspiration (conversion to [mm] per time step)
            # self.var.ESRef = readmapsparse(binding['ES0Maps'], self.var.currentTimeStep(), self.var.ESRef) * self.var.DtDay * self.var.CalEvaporation
            # potential evaporation rate from a bare soil surface (conversion to [mm] per time step)
            self.var.EWRef = readmapsparse(binding['E0Maps'], self.var.currentTimeStep(), self.var.EWRef) * self.var.DtDay * self.var.CalEvaporation
            # potential evaporation rate from water surface (conversion to [mm] per time step)

        self.var.ESRef = (self.var.EWRef + self.var.ETRef)/2

        if option['TemperatureInKelvin']:
            self.var.Tavg -= 273.15



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

from pcraster import*
from pcraster.framework import *

from lisflood.global_modules.add1 import *
from lisflood.global_modules.globals import *


class waterbalance(object):

    """
    # ************************************************************
    # ***** CUMULATIVE MASS BALANCE ERROR*************************
    # ************************************************************
    # Cumulative mass balance error is calculated at outflow point (pit) of
    # each catchment (no pixel values)! In- and outgoing terms below are all
    # cumulative over entire simulation.

    """

    def __init__(self, waterbalance_variable):
        self.var = waterbalance_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the water balance module
        """



        # ************************************************************
        # ***** WATER BALANCE INIT
        # ************************************************************

        # Calculate initial amount of water in catchment (needed to keep track of
        # cumulative mass balance error)

        if (not(option['InitLisflood'])) and option['repMBTs']:
            # not done in Init Lisflood


            ChannelInitM3 = self.var.ChanIniM3
            if not(option['InitLisflood']):    # only with no InitLisflood
                if option['simulateLakes']:
                    ChannelInitM3 += self.var.LakeStorageIniM3
                if option['simulateReservoirs']:
                    ChannelInitM3 += self.var.ReservoirStorageIniM3
                if option['simulatePolders']:
                    ChannelInitM3 += self.var.PolderStorageIniM3

            Hill1 =  self.var.OtherFraction      * (self.var.CumInterception[0] + self.var.W1[0] + self.var.W2[0] + self.var.UZ[0])
            Hill1 += self.var.ForestFraction     * (self.var.CumInterception[1] + self.var.W1[1] + self.var.W2[1] + self.var.UZ[1])
            Hill1 += self.var.IrrigationFraction * (self.var.CumInterception[2] + self.var.W1[2] + self.var.W2[2] + self.var.UZ[2])
            Hill1 += self.var.LZ

            OverlandInitM3 = self.var.OFM3Other + self.var.OFM3Forest + self.var.OFM3Direct

            HillslopeInitM3 = (self.var.SnowCoverInit + Hill1 +
                               self.var.DirectRunoffFraction * self.var.CumInterSealed) * self.var.MMtoM3 + OverlandInitM3
            # Initial water stored at hillslope elements [m3]
            # Note that WInit1, WInit2 and TotalGroundWaterInit are defined for the pixel's permeable fraction
            # only, which is why we need to multiply with PermeableFraction to get the volumes right (no soil moisture
            # or groundwater is stored at all in the direct runoff fraction!)
            # MMtoM3 equals MMtoM*PixelArea, which may (or may not) be
            # spatially variable

            self.var.WaterInit = np.take(np.bincount(self.var.Catchments,
                weights=ChannelInitM3),self.var.Catchments)

            self.var.WaterInit += np.take(np.bincount(self.var.Catchments,
                weights=HillslopeInitM3),self.var.Catchments)

            # Initial water stored [m3]
            # Inclusion of DischargeM3Structures: adding this corrects a (relatively small) offset that occurs otherwise
            # if structures (reservoirs, lakes) are used.
            #DisStructure = cover(ifthen(
            #    self.var.IsUpsOfStructureKinematic, self.var.ChanQ * self.var.DtRouting), scalar(0.0))

            DisStructure = np.where(self.var.IsUpsOfStructureKinematicC,
                        self.var.ChanQ * self.var.DtRouting,0)
#            DisStructure = np.where(self.var.IsUpsOfStructureKinematicC,
#                        self.var.ChanM3,0)

            # Discharge upstream of structure locations (coded as pits) in [m3/time step]
            # Needed for mass balance error calculations (see comment to calculation of WaterInit below)
            # Inclusion of DischargeM3Structures: adding this corrects a (relatively small) offset that occurs otherwise
            # if structures (reservoirs, lakes) are used.
            # calculation of structur influence happens before routing, therefore the initial state is used at the structure onece to often
            # (because it is not routed yet to the structure)

            if option['simulateLakes']:
                DisStructure += np.where(compressArray(self.var.IsUpsOfStructureLake),
                        0.5 * self.var.ChanQ * self.var.DtRouting,0)

             #   DisStructure += cover(ifthen(self.var.IsUpsOfStructureLake,
             #                                0.5 * self.var.ChanQ * self.var.DtRouting), scalar(0.0))
                # because Modified Puls Method is use, some additional offset
                # has to be add
            #self.var.DischargeM3StructuresIni = areatotal(DisStructure, catch)
            self.var.DischargeM3StructuresIni = np.take(np.bincount(self.var.Catchments,
                weights=DisStructure),self.var.Catchments)



          #  report(decompress(self.var.WaterInit),'E:/lisflood_test/newDiepoldsau2/winitPy.map')
# --------------------------------------------------------------------------
# -------------------------------------------------------------------------

    def dynamic(self):
        """ dynamic part of the water balance module
        """

        if (not(option['InitLisflood'])) and option['repMBTs']:

            # ************************************************************
            # ***** CUMULATIVE MASS BALANCE ERROR*************************
            # ************************************************************

                # Cumulative mass balance error is calculated at outflow point (pit) of
                # each catchment (no pixel values)! In- and outgoing terms below are all
                # cumulative over entire simulation.

            # This comes in:
            #WaterIn = areatotal(cover(decompress(self.var.sumIn), scalar(0.0)), catch) + areatotal(
            #    decompress(self.var.TotalPrecipitation) * self.var.MMtoM3, catch)

            self.var.sumIn[np.isnan(self.var.sumIn)] = 0
            WaterIn = np.take(np.bincount(self.var.Catchments,
                weights=self.var.sumIn),self.var.Catchments)
            WaterIn += np.take(np.bincount(self.var.Catchments,
                weights=self.var.TotalPrecipitation*self.var.MMtoM3),self.var.Catchments)


            # WaterIn=areatotal(TotalQInM3,Catchments) + areatotal(TotalPrecipitation*MMtoM3,Catchments);
            # Accumulated incoming water [cu m]
            # NOTE: It is NOT possible to nest all terms into one areatotal statement because channel-related maps have MV for non-channel
            # pixels, resulting in MV creation when adding directly!!
            # MMtoM3 equals MMtoM*PixelArea, which may (or may not) be
            # spatially variable

            # This is stored:
            ChannelStoredM3 = self.var.ChanM3.copy()
            if option['simulateLakes']:
                ChannelStoredM3 += self.var.LakeStorageM3Balance
            if option['simulateReservoirs']:
                ChannelStoredM3 += self.var.ReservoirStorageM3
            if option['simulatePolders']:
                ChannelStoredM3 += self.var.PolderStorageM3

            # ChannelStoredM3=ChanM3 + ReservoirStorageM3 + LakeStorageM3 + PolderStorageM3;
            # Water stored in channel network [m3] (including reservoirs,
            # lakes, polders)

            Hill1 = self.var.OtherFraction * (self.var.CumInterception[0] + self.var.W1[0] + self.var.W2[0] + self.var.UZ[0])
            Hill1 += self.var.ForestFraction * (self.var.CumInterception[1] + self.var.W1[1] + self.var.W2[1] + self.var.UZ[1])
            Hill1 += self.var.IrrigationFraction * (self.var.CumInterception[2] + self.var.W1[2] + self.var.W2[2] + self.var.UZ[2]) + self.var.LZ

            HillslopeStoredM3 = (self.var.WaterDepth + self.var.SnowCover + Hill1 +
                                 self.var.DirectRunoffFraction * self.var.CumInterSealed) * self.var.MMtoM * self.var.PixelArea
            # Water stored at hillslope elements [m3]
            # Note that W1, W2 and TotalGroundWater are defined for the pixel's permeable fraction
            # only, which is why we need to multiply with PermeableFraction to get the volumes right (no soil moisture
            # or groundwater is stored at all in the direct runoff fraction!)

            # OverlandM3 = self.var.OFM3Other + self.var.OFM3Forest + self.var.OFM3Direct

            #WaterStored = areatotal(decompress(ChannelStoredM3), catch) + areatotal(decompress(HillslopeStoredM3), catch)
            WaterStored = np.take(np.bincount(self.var.Catchments,weights=ChannelStoredM3),self.var.Catchments)
            WaterStored += np.take(np.bincount(self.var.Catchments,weights=HillslopeStoredM3),self.var.Catchments)

            # Total water stored [m3]

            # This goes out:
            HillslopeOutM3 = (self.var.TaCUM + self.var.TaInterceptionCUM +
                              self.var.ESActCUM + self.var.GwLossCUM) * self.var.MMtoM3
            # Water that goes out of the system at the hillslope level [m3]
            # (evaporation and groundwater loss)

            sum1 = self.var.sumDis.copy()
            sum1[self.var.AtLastPointC == 0] = 0
            WaterOut = np.take(np.bincount(self.var.Catchments,weights=sum1 * self.var.DtRouting),self.var.Catchments)
            WaterOut += np.take(np.bincount(self.var.Catchments,weights=HillslopeOutM3),self.var.Catchments)

            #WaterOut = areatotal(cover(ifthen(
            #    self.var.AtLastPoint, self.var.sumDis * self.var.DtRouting), scalar(0.0)), catch)
            #WaterOut += areatotal(decompress(HillslopeOutM3), catch)
            if option['simulateLakes']:
                WaterOut += np.take(np.bincount(self.var.Catchments,
                    weights=self.var.EWLakeCUMM3),self.var.Catchments)
            if option['openwaterevapo']:
                WaterOut += np.take(np.bincount(self.var.Catchments,
                    weights=self.var.EvaCumM3),self.var.Catchments)
            if option['TransLoss']:
                WaterOut += np.take(np.bincount(self.var.Catchments,
                    weights=self.var.TransCum),self.var.Catchments)
            if option['wateruse']:
                WaterOut += np.take(np.bincount(self.var.Catchments,
                    weights=self.var.IrriLossCUM),self.var.Catchments)
                WaterOut += np.take(np.bincount(self.var.Catchments,
                    weights=self.var.wateruseCum),self.var.Catchments)
            # Accumulated outgoing water [cu m]
            # Inclusion of DischargeM3Structures is because at structure locations the water in the channel is added to the structure
            # (i.e. storage at reservoirs/lakes is accounted for twice). Of course this is not really a 'loss', but merely a correction
            # for this double counting)
            # new 12.11.09 PB
            # added cumulative transmission loss
            DisStru = np.where(self.var.IsUpsOfStructureKinematicC,
                        self.var.ChanQ * self.var.DtRouting,0)
            DischargeM3Structures = np.take(np.bincount(self.var.Catchments,
                weights=DisStru),self.var.Catchments)


            #DischargeM3Structures = areatotal(cover(ifthen(
            #    self.var.IsUpsOfStructureKinematic, self.var.ChanQ * self.var.DtRouting), scalar(0.0)), catch)
            # on the last time step lakes and reservoirs calculated with the previous routing results
            # so the last (now routed) discharge has to be added to the mass balance
            # (-> the calculation odf the structures is done before the routing)

            if option['simulateLakes']:
                DisLake = globals.inZero.copy()
                np.put(DisLake,self.var.LakeIndex,0.5 * self.var.LakeInflowCC * self.var.DtRouting)
                DischargeM3Lake = np.take(np.bincount(self.var.Catchments,
                    weights=DisLake),self.var.Catchments)
                #DischargeM3Lake = areatotal(cover(0.5 * self.var.LakeInflow * self.var.DtRouting, scalar(0.0)), catch)
                # because Modified Puls Method is using QIn=(Qin1+Qin2)/2, we need a correction
                #  DisStr=Disstr+0.5*LakeInflow - 0.5 * LakeInit
                #  0.5 * LakeInit: is already done in DischargeM3StructuresIni
                DischargeM3Structures += DischargeM3Lake
                # Discharge just upstream of structure locations (coded as pits) in [cu m / time step]
                # Needed for mass balance error calculations, because of double counting of structure
                # storage and water in the channel.

            DischargeM3Structures -= self.var.DischargeM3StructuresIni
            # minus the initial DischargeStructure

            # Old: DischargeM3Structures=areatotal(cover(ifthen(self.var.IsUpsOfStructureKinematic,self.var.ChanQ*self.var.DtSec),null),self.var.Catchments)
            # Discharge just upstream of structure locations (coded as pits) in [cu m / time step]
            # Needed for mass balance error calculations, because of double counting of structure
            # storage and water in the channel.

            # Mass balance:
            #  self.var.MBError = self.var.WaterInit + WaterIn - WaterStored - WaterOut - DischargeM3Structures
            self.var.MBError = self.var.WaterInit + WaterIn - WaterStored - WaterOut - \
                               DischargeM3Structures
            # Cumulative mass balance error per catchment [cu m]

            #CatchArea = areatotal(decompress(self.var.PixelAreaArray), catch)
            CatchArea = np.take(np.bincount(self.var.Catchments,
                    weights=self.var.PixelArea),self.var.Catchments)

            self.var.MBErrorMM = self.var.MtoMM * self.var.MBError / CatchArea

            #self.var.MBError = decompress(self.var.MBError)
            #self.var.MBErrorMM = decompress(self.var.MBErrorMM)
            # self.var.MBErrorMM = self.var.MtoMM * self.var.MBError * self.var.InvCatchArea
            # expressed as mm water slice



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""

import xml.dom.minidom
import datetime
import time as xtime
import os
from netCDF4 import Dataset, date2num, num2date

try:
    from netCDF4 import netcdftime
except ImportError:
    import cftime as netcdftime  # newer versions of netCDF4 don't include netcdftime

from dateutil import parser
from platform import system as operating_system
from pandas._libs.tslibs.parsing import parse_time_string

from pcraster import*
from pcraster.framework import *

from globals import *

import numpy as np
MAX_READ_TRIALS = 100 # max number of trial allowed re-read an input file: to avoid crashes due to temporary network interruptions
READ_PAUSE = 0.1      # pause (seconds) between each re-read trial over the network
try:
    NC_DATE_TYPE = netcdftime._netcdftime.datetime
except:
    NC_DATE_TYPE = netcdftime.datetime  # work-around for older versions of the netCDF4 library (< 1.2.7 ?)


project_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '../..'))


class LisfloodError(Exception):
    """
    the error handling class
    prints out an error
    """
    def __init__(self, msg):
        header = "\n\n ========================== LISFLOOD ERROR =============================\n"
        try:
           self._msg = header + msg +"\n" +  sys.exc_info()[1].message
        except:
           self._msg = header + msg +"\n"

    def __str__(self):
        return self._msg


class LisfloodFileError(LisfloodError):
    """
    the error handling class
    prints out an error
    """
    def __init__(self, filename,msg=""):
        path, name = os.path.split(filename)
        path = os.path.normpath(path)
        if os.path.exists(path):
            text1 = "path: " + path + " exists\nbut filename: " + name + " does not\n"
            text1 += "file name extension can be .map or .nc\n"
        else:
            text1 = "searching: " + filename
            text1 += "\npath: " + path + " does not exists\n"

        header = "\n\n ======================== LISFLOOD FILE ERROR ===========================\n"
        self._msg = header + msg + text1


class LisfloodWarning(Warning):
    """
    the error handling class
    prints out an error
    """
    def __init__(self, msg):
        header = "\n\n ========================== LISFLOOD Warning =============================\n"
        self._msg = header + msg

    def __str__(self):
        return self._msg


class LisfloodRunInfo(Warning):
    """
    the error handling class
    prints out an error
    """
    def __init__(self, mode, outputDir, Steps = 1, ensMembers=1, Cores=1):
        header = "\n\n ========================== LISFLOOD Simulation Information and Setting =============================\n"
        msg = "   LISFLOOD is used in the "+str(mode)+"\n"
        if ensMembers > 1:
            msg += "   It uses "+str(ensMembers)+" ensemble members for the simulation\n"
        if Steps > 1:
            msg += "   The model will be updated at "+str(Steps)+" time step during the simulation\n"
        if Cores > 1:
            msg += "   The simulation will try to use "+str(Cores)+" processors simultaneous\n"
        msg += "   The simulation output as specified in the settings file can be found in "+str(outputDir)+"\n"
        self._msg = header + msg

    def __str__(self):
        return self._msg


def optionBinding(settingsfile, optionxml):
    """ Read XML settings and options files 
    
    Read Settings.XML and Options.XML input files and return:
    binding = key and value (filename or value)
    option  = control of Lisflood to use certain subroutines
    ReportSteps = maps are reported at these time steps
    
    :param settingsfile: path and name of Settings.xml
    :param optionxml: path and name of Options.xml
    :return: binding, option, ReportSteps (as global variables)
    """
    settings_path = os.path.normpath(os.path.abspath(os.path.dirname(settingsfile)))
    optionSetting = {}

    # user defined path, parameters, calibration parameters, dates, ect from settingsxml
    # we add here some built-in user variables
    user = {
        'ProjectDir': project_dir, 'ProjectPath': project_dir,
        'SettingsDir': settings_path, 'SettingsPath': settings_path,
    }
    repTimeserie = {}
    repMaps = {}

    # domopt = xml.dom.minidom.parseString(optionxml)

    #read xml optionxml i.e. OptionTserieMaps.xml file
    try:
        f=open(optionxml)
        f.close()
    except:
        msg = "Cannot find option file: " + optionxml
        raise LisfloodFileError(optionxml,msg)
    try:
        #DOM object representing content of optionxml file
        domopt = xml.dom.minidom.parse(optionxml)
    except:
        msg = "Error using option file: " + optionxml
        raise LisfloodError(msg)

    #read file settings (settingsfile)
    try:
       f = open(settingsfile,'r')
       xmlstring1 = ""
       while 1:
          line = f.readline()
          if not line:break
          xmlstring1 += line
       f.close()
       xmlstring.append(xmlstring1)
    except:
        msg = "Cannot find settings file: " + settingsfile
        raise LisfloodFileError(settingsfile,msg)
    try:
        #DOM object representing content of settings file
        dom = xml.dom.minidom.parse(settingsfile)
    except:
        msg = "Error using: " + settingsfile
        raise LisfloodError(msg)

    # getting all posssible option from the general optionxml
    # and setting them to their default value
    # list of elements "lfoptions" in optionxml file
    optDef = domopt.getElementsByTagName("lfoptions")[0]
    for optset in optDef.getElementsByTagName("setoption"):
        option[optset.attributes['name'].value] = bool(
            int(optset.attributes['default'].value))

    # getting option set in the specific settings file
    # and resetting them to their choice value
    # list of elements "lfoptions" in settings file
    optSet = dom.getElementsByTagName("lfoptions")[0]
    for optset in optSet.getElementsByTagName("setoption"):
        optionSetting[optset.attributes['name'].value] = bool(
            int(optset.attributes['choice'].value))
    for key in optionSetting.keys():
        # set the same value for corresponding keys, settings overwrites optionxml
        option[key] = optionSetting[key]

    # reverse the initLisflood option to use it as a restriction for output
    # eg. produce output if not(initLisflood)
    option['nonInit'] = not(option['InitLisflood'])
# -----------------------------------------

    # get all the bindings in the first part of the settingsfile = lfuser
    # list of elements "lfuser" in settings file
    lfuse = dom.getElementsByTagName("lfuser")[0]
    for userset in lfuse.getElementsByTagName("textvar"):
        user[userset.attributes['name'].value] = str(userset.attributes['value'].value)

    # get all the binding in the last part of the settingsfile  = lfbinding
    # list of elements "lfbinding" in settings file
    bind = dom.getElementsByTagName("lfbinding")[0]
    for bindset in bind.getElementsByTagName("textvar"):
        binding[bindset.attributes['name'].value] = str(bindset.attributes['value'].value)

    # replace/add the information from lfuser to lfbinding
    for i in binding.keys():
        expr = binding[i]
        while expr.find('$(') > -1:
            a1 = expr.find('$(')
            a2 = expr.find(')')
            try:
                s2 = user[expr[a1 + 2:a2]]
            except KeyError:
                print 'no ', expr[a1 + 2:a2],'for',binding[i] ,' in lfuser defined'
            expr = expr.replace(expr[a1:a2 + 1], s2)
        binding[i] = expr

    # if pathout has some placeholders, they are replace here
    pathout = user["PathOut"]
    while pathout.find('$(') > -1:
        a1 = pathout.find('$(')
        a2 = pathout.find(')')
        try:
            s2 = user[pathout[a1 + 2:a2]]
        except KeyError:
            print 'no ', expr[a1 + 2:a2],'for',pathout,' in lfuser defined'
        pathout = pathout.replace(pathout[a1:a2 + 1], s2)

    #CM: output folder
    outputDir.append(pathout)
    #outputDir.append(user["PathOut"])

    # Read the calendar type from the precipitation forcing NetCDF file
    with Dataset(binding["PrecipitationMaps"] + ".nc") as nc:
        binding["calendar_type"] = getCalendarType(nc)

# ---------------------------------------------
    # Split the string ReportSteps into an int array
    # replace endtime with number
    # replace .. with sequence

    try:
        repsteps = user['ReportSteps'].split(',')
    except:
        repsteps = 'endtime'
    if repsteps[-1] == 'endtime':
        repsteps[-1] = binding['StepEnd']
    jjj = []
    for i in repsteps:
        if '..' in i:
            j = map(int, i.split('..'))
            for jj in xrange(j[0], j[1] + 1):
                jjj.append(jj)
        else:
            jjj.append(i)
    ReportSteps['rep'] = map(datetoInt, jjj)
    # maps are reported at these time steps

# -----------------------------------------------

    # Get output Directory from settings file
    try:
        nrCores.append(int(user["nrCores"]))
    except:
        nrCores.append(int(1))

# -------------------------------------------

    # Split the string FilterSteps into an int array
    # remove endtime if present
    # replace .. with sequence

    try:
        filterSteps = user['FilterSteps'].split(',')
    except:
        filterSteps = int(0)
    #try:
    #    filterSteps = user['FilterSteps'].split(' ')
    #except:
    #    filterSteps = int(0)
    if filterSteps[-1] == 'endtime' or filterSteps[-1] == binding['StepEnd'] or filterSteps[-1] == binding['StepEnd']:
        filterSteps[-1] = int(0)
    for i in filterSteps:
        try:
            timeDif = datetime.datetime.strptime(i, "%d/%m/%Y") - datetime.datetime.strptime(binding['CalendarDayStart'], "%d/%m/%Y")
            val = int(timeDif.days)
        except:
            val = int(i)
        stependint = datetoInt(binding['StepEnd'])
        #if int(val) < int(binding['StepEnd']):
        if int(val) < stependint:
            try:
                FilterSteps.append(int(i))
            except:
                FilterSteps.append(int(timeDif.days))
    # maps are reported at these time steps

#----------------------------------------------

    # Number of Ensemble members for MonteCarlo or EnKF
    try:
        EnsMembers.append(int(user["EnsMembers"]))
    except:
        EnsMembers.append(int(1))

    # Number of cores for MonteCarlo or EnKF
    try:
        nrCores.append(int(user["nrCores"]))
    except:
        nrCores.append(int(1))

# -------------------------
    # running through all times series
    # list of elements "lftime" in optionxml file
    reportTimeSerie = domopt.getElementsByTagName("lftime")[0]
    for repTime in reportTimeSerie.getElementsByTagName("setserie"):
        d = {}
        for key in repTime.attributes.keys():
            if key != 'name':
                value = repTime.attributes[key].value
                d[key] = value.split(',')
        key = repTime.attributes['name'].value
        repTimeserie[key] = d
        repOpt = repTimeserie[key]['repoption']
        try:
            restOpt = repTimeserie[key]['restrictoption']
        except:
            # add restricted option if not in already
            repTimeserie[key]['restrictoption'] = ['']
            restOpt = repTimeserie[key]['restrictoption']
        try:
            test = repTimeserie[key]['operation']
        except:
            # add operation if not in already
            repTimeserie[key]['operation'] = ['']

        # sort out if this option is not active
        # put in if one of this option is active
        for i in repOpt:
            for o1key in option.keys():
                if option[o1key]:  # if option is active = 1
                    # print o1key, option[o1key],i
                    if o1key == i:
                        # option is active and time series has this option to select it
                        # now test if there is any restrictions
                        allow = True
                        for j in restOpt:
                            for o2key in option.keys():
                                if o2key == j:
                                    # print o2key, option[o2key],j
                                    if not(option[o2key]):
                                        allow = False
                        if allow:
                            reportTimeSerieAct[key] = repTimeserie[key]

# -------------------------
    # running through all maps

    reportMap = domopt.getElementsByTagName("lfmaps")[0]

    for repMap in reportMap.getElementsByTagName("setmap"):
        d = {}
        for key in repMap.attributes.keys():
            if key != 'name':
                value = repMap.attributes[key].value
                d[key] = value.split(',')
        key = repMap.attributes['name'].value
        repMaps[key] = d
        try:
            repAll = repMaps[key]['all']
        except:
            repMaps[key]['all'] = ['']
            repAll = ['']
        try:
            repSteps = repMaps[key]['steps']
        except:
            repMaps[key]['steps'] = ['']
            repSteps = ['']
        try:
            repEnd = repMaps[key]['end']
        except:
            repMaps[key]['end'] = ['']
            repEnd = ['']
        try:
            restOpt = repMaps[key]['restrictoption']
        except:
            # add restricted option if not in already
            repMaps[key]['restrictoption'] = ['']
            restOpt = repMaps[key]['restrictoption']
        try:
            repUnit = repMaps[key]['unit']
        except:
            repMaps[key]['unit'] = ['-']
        #  -------- All -----------------
        # sort out if this option is not active
        # put in if one of this option is active
        for i in repAll:
            # run through all the output option
            for o1key in option.keys():
                # run through all the options
                if option[o1key]:  # if option is active = 1
                    # print o1key, option[o1key],i
                    if o1key == i:
                        # option is active and time series has this option to select it
                        # now test if there is any restrictions
                        allow = True
                        for j in restOpt:
                            # running through all the restrictions
                            for o2key in option.keys():
                                if (o2key == j) and (not(option[o2key])):
                                    allow = False
                        if allow:
                            reportMapsAll[key] = repMaps[key]

        #  -------- Steps -----------------
        for i in repSteps:
            for o1key in option.keys():
                if option[o1key]:  # if option is active = 1
                    if o1key == i:
                        allow = True
                        for j in restOpt:
                            for o2key in option.keys():
                                if (o2key == j) and (not(option[o2key])):
                                    allow = False
                        if allow:
                            reportMapsSteps[key] = repMaps[key]

        #  -------- End -----------------
        for i in repEnd:
            for o1key in option.keys():
                if option[o1key]:  # if option is active = 1
                    if o1key == i:
                        allow = True
                        for j in restOpt:
                            for o2key in option.keys():
                                if (o2key == j) and (not(option[o2key])):
                                    allow = False
                        if allow:
                            reportMapsEnd[key] = repMaps[key]

    # return option,binding,ReportSteps
    # return option,ReportSteps
    # return ReportSteps
    return


def counted(fn):
    def wrapper(*args, **kwargs):
        wrapper.called+= 1
        return fn(*args, **kwargs)
    wrapper.called= 0
    wrapper.__name__= fn.__name__
    return wrapper

@counted
def checkmap(name, value, map, flagmap, find):
    """ Check if maps fit to the mask map
    
    :param name: key in Settings.xml containing path and name of the map to be checked as string (map name)
    :param value: name and path of the map to be checked as string
    :param map: map to be used for checking in pcraster format
    :param flagmap: flag for maps
    :param find: 
    :return: 
    """
    s = [name, value]
    if flagmap:
        amap = scalar(defined(MMaskMap))
        try:
            smap = scalar(defined(map))
        except:
            msg = "Map: " + name + " in " + value + " does not fit"
            if name == "LZAvInflowMap":
                msg +="\nMaybe run initial run first"
            raise LisfloodError(msg)

        mvmap = maptotal(smap)
        mv = cellvalue(mvmap, 1, 1)[0]
        s.append(mv)

        less = maptotal(ifthenelse(defined(MMaskMap), amap - smap, scalar(0)))
        s.append(cellvalue(less, 1, 1)[0])
        less = mapminimum(scalar(map))
        s.append(cellvalue(less, 1, 1)[0])
        less = maptotal(scalar(map))
        if mv > 0:
            s.append(cellvalue(less, 1, 1)[0] / mv)
        else:
            s.append('0')
        less = mapmaximum(scalar(map))
        s.append(cellvalue(less, 1, 1)[0])
        if find > 0:
            if find == 2:
                s.append('last_Map_used')
            else:
                s.append('')

    else:
        s.append(0)
        s.append(0)
        s.append(float(map))
        s.append(float(map))
        s.append(float(map))

    # print check results
    if checkmap.called == 1:
        print "%-25s%-40s%11s%11s%11s%11s%11s" %("Name","File/Value","nonMV","MV","min","mean","max")
    print "%-25s%-40s%11i%11i%11.2f%11.2f%11.2f" %(s[0],s[1][-39:],s[2],s[3],s[4],s[5],s[6])
    return


def timemeasure(name,loops=0, update = False, sample = 1):
    # returns the current processor time
    timeMes.append(xtime.clock())
    if loops == 0:
        s = name
    else:
        s = name+"_%i" %(loops)
    timeMesString.append(s)
    return


def Calendar(input):
    """ Get date or number of steps from input.

    Get date from input string using one of the available formats or get time step number from input number or string.
    Used to get the date from CalendarDayStart (input) in the settings xml

    :param input: string containing a date in one of the available formats or time step number as number or string
    :rtype: datetime object or float number
    :returns: date as datetime or time step number as float 
    :raises ValueError: stop if input is not a step number AND it is in wrong date format
    """

    try:
        # try reading step number from number or string
        return float(input)
    except:
        # try reading a date in one of available formats
        try:
            _t_units = "hours since 1970-01-01 00:00:00" # units used for date type conversion (datetime.datetime -> calendar-specific if needed)
            date = parse_time_string(input, dayfirst=True)[0]  # datetime.datetime type
            step = date2num(date, _t_units, binding["calendar_type"]) # float type
            return num2date(step, _t_units, binding["calendar_type"]) # calendar-dependent type from netCDF4.netcdftime._netcdftime module
        except:    
            # if cannot read input then stop
            msg = "Wrong step or date format in XML settings file\n" \
                  "Input " + str(input)
            raise LisfloodError(msg)


def datetoInt(dateIn,both=False):
    """ Get number of steps between dateIn and CalendarDayStart.
    
    Get the number of steps between dateIn and CalendarDayStart and return it as integer number.
    It can now compute the number of sub-daily steps.
    dateIn can be either a date or a number. If dateIn is a number, it must be the number of steps between
    dateIn and CalendarDayStart.
    
    :param dateIn: date as string or number
    :param both: if true it returns both the number of steps as integer and the input date as string. If false only
    the number of steps as integer is returned
    :return: number of steps as integer and input date as string
    """

    # get reference date to be used with step numbers from 'CalendarDayStart' in Settings.xml file
    date1 = Calendar(dateIn)
    begin = Calendar(binding['CalendarDayStart'])
    # get model time step as float form 'DtSec' in Settings.xml file
    DtSec = float(binding['DtSec'])
    # compute fraction of day corresponding to model time step as float
    DtDay = float(DtSec / 86400.)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    if type(date1) is datetime.datetime:
         str1 = date1.strftime("%d/%m/%Y %H:%M")
         # get total number of seconds corresponding to the time interval between dateIn and CalendarDayStart
         timeinterval_in_sec = int((date1 - begin).total_seconds())
         # get total number of steps between dateIn and CalendarDayStart
         int1 = int(timeinterval_in_sec/DtSec +1)
         # int1 = (date1 - begin).days + 1
    else:
        int1 = int(date1)
        str1 = str(date1)
    if both: return int1,str1
    else: return int1


def inttoDate(intIn,refDate):
    """ Get date corresponding to a number of steps from a reference date.

    Get date corresponding to a number of steps from a reference date and return it as datetime.
    It can now use sub-daily steps.
    intIn is a number of steps from the reference date refDate.

    :param intIn: number of steps as integer
    :param refDate: reference date as datetime
    :return: stepDate: date as datetime corresponding to intIn steps from refDate
    """

    # CM: get model time step as float form 'DtSec' in Settings.xml file
    DtSec = float(binding['DtSec'])
    # CM: compute fraction of day corresponding to model time step as float
    DtDay = float(DtSec / 86400)
    # Time step, expressed as fraction of day (same as self.var.DtSec and self.var.DtDay)

    # CM: compute date corresponding to intIn steps from reference date refDate
    stepDate = refDate + datetime.timedelta(days=(intIn*DtDay))

    return stepDate


def checkifDate(start,end):
    """ Check simulation start and end dates or timesteps
    
    Check simulation start and end dates/timesteps to be later than begin date (CalendarStartDay).
    If dates are used for binding[start] and binding[end], it substitutes dates with time step numbers.
    
    :param start: start date for model run (# or date as string)
    :param end: end date for model run (# or date as string)
    :returns: modelSteps (modelSteps[0] = intStart 
    modelSteps.append(intEnd)
    """
    # calendar date start (CalendarDayStart)
    begin = Calendar(binding['CalendarDayStart'])

    intStart,strStart = datetoInt(binding[start],True)
    # overwrite date with time step
    binding[start] = intStart
    intEnd,strEnd = datetoInt(binding[end],True)
    binding[end] = intEnd

    # test if start and end > begin
    if (intStart<0) or (intEnd<0) or ((intEnd-intStart)<0):
        strBegin = begin.strftime("%d/%m/%Y %H:%M")
        msg="Simulation start date and/or simulation end date are wrong or do not match CalendarStartDate!\n"+ \
            "CalendarStartDay: "+strBegin +"\n" + \
            "Simulation start: "+strStart + " - "+str(intStart)+"\n" + \
            "Simulation end: "+strEnd + " - "+str(intEnd)
        raise LisfloodError(msg)
    modelSteps.append(intStart)
    modelSteps.append(intEnd)
    return


class DynamicFramework(DynamicFramework):
    """
    Framework class for dynamic models.
    `userModel`
    Instance that models the :ref:`Dynamic Model Concept <dynamicModelConcept>`.
    `lastTimeStep`   Last timestep to run.
    `firstTimestep`  Sets the starting timestep of the model (optional, default is 1).
    Updated by zusatz.py
    """

    def run(self):
        """
        Run the dynamic user model.

        .. todo::
        This method depends on the filter frameworks concept. Shouldn't its run
        method call _runSuspend()?
        """
        self._atStartOfScript()
        if(hasattr(self._userModel(), "resume")):
            #if self._userModel().firstTimeStep() == 1:
            # replaced this because starting date is not always the 1
            if self._userModel().firstTimeStep() == datetoInt(binding['StepStart']):
               self._runInitial()
            else:
               self._runResume()
        else:
            self._runInitial()

        self._runDynamic()

        # Only execute this section while running filter frameworks.
        if hasattr(self._userModel(), "suspend") and \
        hasattr(self._userModel(), "filterPeriod"):
          self._runSuspend()

        return 0




    """Adjusting the def _atStartOfTimeStep defined in DynamicFramework
       for a real quiet output
    """
    rquiet = False
    rtrace = False

    def _atStartOfTimeStep(self, step):
        self._userModel()._setInTimeStep(True)
        if not self.rquiet:
            if not self.rtrace:
                msg = u"#"

            else:
                msg = u"%s<time step=\"%s\">\n" % (self._indentLevel(), step)
            sys.stdout.write(msg)
            sys.stdout.flush()


class EnsKalmanFilterFramework(EnsKalmanFilterFramework):
  """
   Framework for particle filter runs
   Updated by zusatz
  """
  def _startEndOfPeriod(self, currentPeriod, lastPeriod):
    # determine start end end timestep of current period
    if currentPeriod == 0:
      #startTimestep = 1  # use first timestep instead of hardcode 1
      startTimestep = self._userModel().firstTimeStep()

      endTimestep = self._userModel()._d_filterTimesteps[currentPeriod]
    elif currentPeriod == lastPeriod:
      startTimestep = self._userModel()._d_filterTimesteps[currentPeriod -1] + 1
      endTimestep = self._d_totalTimesteps
    else:
      startTimestep = self._userModel()._d_filterTimesteps[currentPeriod - 1] + 1
      endTimestep = self._userModel()._d_filterTimesteps[currentPeriod]

    assert startTimestep <= endTimestep
    return startTimestep, endTimestep


class TimeoutputTimeseries(TimeoutputTimeseries):
    """
    Class to create pcrcalc timeoutput style timeseries
    Updated py zusatz.py
    """
    def _writeFileHeader(self, outputFilename):
        """
        writes header part of tss file
        """
        outputFile = open(outputFilename, "w")
        # header
        #outputFile.write("timeseries " + self._spatialDatatype.lower() + "\n")
        outputFile.write("timeseries " + self._spatialDatatype.lower() + " settingsfile: "+os.path.realpath(sys.argv[1])+" date: " + xtime.ctime(xtime.time())+ "\n")
        sys.argv[1]
        # write number of outlets points +1
        outputFile.write(str(self._ncodesId + 1) + "\n")
        outputFile.write("timestep\n")

        for colId in range(0, self._ncodesId):
            outputFile.write(str(self._codesId[colId]) + "\n")
        outputFile.close()

    def _configureOutputFilename(self, filename):
        """
        generates filename
        appends timeseries file extension if necessary
        prepends sample directory if used in stochastic
        """

        # test if suffix or path is given
        head, tail = os.path.split(filename)

        if not re.search("\.tss", tail):
            # content,sep,comment = filename.partition("-")
            # filename = content + "Tss" + sep + comment + ".tss"
            filename = tail + ".tss"

        # for stochastic add sample directory
        if hasattr(self._userModel, "nrSamples"):
            try:
                filename = os.path.join(str(self._userModel.currentSampleNumber()), filename)
            except:
                pass

        return filename

    def _writeTssFile(self):
        """
        writing timeseries to disk
        """
        #
        outputFilename =  self._configureOutputFilename(self._outputFilename)
        a= option['EnKF']
        outputFile = None
        if option['EnKF']:
            if os.path.exists(outputFilename) == False:
               if self._writeHeader == True:
                  self._writeFileHeader(outputFilename)
                  outputFile = open(outputFilename, "a")
               else:
                  outputFile = open(outputFilename, "w")
            else:
                outputFile = open(outputFilename, "a")
        else:
            if self._writeHeader == True:
                  self._writeFileHeader(outputFilename)
                  outputFile = open(outputFilename, "a")
            else:
                  outputFile = open(outputFilename, "w")


        assert outputFile

        start = self._userModel.firstTimeStep()
        end = self._userModel.nrTimeSteps() + 1

        for timestep in range(start, end):
            row = ""
            row += " %8g" % timestep
            if self._spatialIdGiven:
                for cellId in range(0, self._ncodesId):
                    value = self._sampleValues[timestep - start][cellId]
                    if isinstance(value, Decimal):
                        row += "           1e31"
                    else:
                        row += " %14g" % (value)
                row += "\n"
            else:
                value = self._sampleValues[timestep - start]
                if isinstance(value, Decimal):
                    row += "           1e31"
                else:
                    row += " %14g" % (value)
                row += "\n"

            outputFile.write(row)

        outputFile.close()

    def __init__(self, tssFilename, model, idMap=None, noHeader=False):
        """

        """

        if not isinstance(tssFilename, str):
            raise Exception(
                "timeseries output filename must be of type string")

        self._outputFilename = tssFilename
        self._maxId = 1
        self._ncodesId = 1
        self._spatialId = None
        self._spatialDatatype = None
        self._spatialIdGiven = False
        self._userModel = model
        self._writeHeader = not noHeader
        # array to store the timestep values
        self._sampleValues = None

        _idMap = False
        if isinstance(idMap, str) or isinstance(idMap, pcraster._pcraster.Field):
            _idMap = True

        #nrRows = 10000 - self._userModel.firstTimeStep() + 1
        #nrRows = self._userModel.nrTimeSteps() - self._userModel.firstTimeStep() + 1

        # if header reserve rows from 1 to endstep
        # if noheader only from startstep - endstep
        #if noHeader:
        #    nrRows = int(binding['StepEnd']) - int(binding['StepStart']) - self._userModel.firstTimeStep() + 2
        #else: nrRows = int(binding['StepEnd']) - int(binding['StepStart']) - self._userModel.firstTimeStep() + 2
        if noHeader:
            nrRows = datetoInt(binding['StepEnd']) - datetoInt(binding['StepStart']) - self._userModel.firstTimeStep() + 2
        else: nrRows = datetoInt(binding['StepEnd']) - datetoInt(binding['StepStart']) - self._userModel.firstTimeStep() + 2



        if _idMap:
            self._spatialId = idMap
            if isinstance(idMap, str):
                self._spatialId = iterReadPCRasterMap(idMap)

            _allowdDataTypes = [
                pcraster.Nominal, pcraster.Ordinal, pcraster.Boolean]
            if self._spatialId.dataType() not in _allowdDataTypes:
                #raise Exception(
                #    "idMap must be of type Nominal, Ordinal or Boolean")
                # changed into creating a nominal map instead of bailing out
                self._spatialId = pcraster.nominal(self._spatialId)

            if self._spatialId.isSpatial():
                self._maxId, valid = pcraster.cellvalue(
                    pcraster.mapmaximum(pcraster.ordinal(self._spatialId)), 1)
                # convert to numpy array
                outletsmapnp = pcr2numpy(self._spatialId,np.nan)
                # get outlets codes from outlets map
                codesId = numpy.unique(outletsmapnp)
                # drop negative values (= missing data in pcraster map)
                codesId = codesId[codesId > 0]
                # get number of outlets points
                self._ncodesId = len(codesId)
                # prepare array to store outlets codes
                self._codesId = [-9999 for i in xrange(self._ncodesId)]

            else:
                self._maxId = 1
                self._codesId = [1]
                self._ncodesId = len(self._codesId)

            # cell indices of the sample locations

            # #self._sampleAddresses = []
            # for cellId in range(1, self._maxId + 1):
            # self._sampleAddresses.append(self._getIndex(cellId))


            # prepare array to store outlets points raster numbers
            self._sampleAddresses = [-9999 for i in xrange(self._ncodesId)]
            # init with the left/top cell - could also be 0 but then you have to catch it in
            # the sample routine and put an exeption in
           # number of cells in map
            nrCells = pcraster.clone().nrRows() * pcraster.clone().nrCols()
            for cell in xrange(1, nrCells + 1):
                if (pcraster.cellvalue(self._spatialId, cell)[1]):
                    # get point code from outlets map for pixel cell
                    outlet_code = pcraster.cellvalue(self._spatialId, cell)[0]
                    # get index of the point code in the sorted list of outlets codes
                    outlet_idx = np.where(codesId == outlet_code)[0][0]
                    # store point code
                    self._codesId[outlet_idx] = outlet_code
                    # store outlets location (cell)
                    self._sampleAddresses[outlet_idx] = cell

            self._spatialIdGiven = True

            nrCols = self._ncodesId
            self._sampleValues = [
                [Decimal("NaN")] * nrCols for _ in [0] * nrRows]
        else:
            self._sampleValues = [[Decimal("NaN")] * 1 for _ in [0] * nrRows]

    def firstout(self,expression):
        """
        returns the first cell as output value
        """
        try:
            cellIndex = self._sampleAddresses[0]
            tmp = pcraster.areaaverage(pcraster.spatial(expression), pcraster.spatial(self._spatialId))
            value, valid = pcraster.cellvalue(tmp, cellIndex)
            if not valid:
               value = Decimal("NaN")
        except:
            value = Decimal("NaN")
        return value



#####################################################################################################################
# TOOLS TO OPEN/READ INPUT FILES ITERATIVELY, IN CASE OF NETWORK TEMPORARILY DOWN
#####################################################################################################################

def iterOpenNetcdf(file_path, error_msg, mode, **kwargs):
    """Wrapper around netCDF4.Dataset function exploiting the iterAccess class to access file_path according to the specified mode"""
    access_function = lambda path: Dataset(path, mode, **kwargs)
    return remoteInputAccess(access_function, file_path, error_msg)


def iterReadPCRasterMap(file_path, error_msg=""):
    """Wrapper around pcraster.readmap function exploiting the iterAccess class to open file_path."""
    return remoteInputAccess(pcraster.readmap, file_path, error_msg)


def iterSetClonePCR(file_path, error_msg=""):
    """Wrapper around pcraster.setclone function exploiting the iterAccess class to access file_path."""
    return remoteInputAccess(pcraster.setclone, file_path, error_msg)


def remoteInputAccess(function, file_path, error_msg):
    """
    Wrapper around the provided file access function.
    It allows re-trying the open/read operation if network is temporarily down.
    Arguments:
        function: function to be called to read/open the file.
        file_path: path of the file to be read/open.
    """
    num_trials = 1
    bad_sep = "/" if operating_system() == "Windows" else "\\"
    file_path = file_path.replace(bad_sep, os.path.sep)
    root = os.path.sep.join(file_path.split(os.path.sep)[:4])
    while num_trials <= MAX_READ_TRIALS:
        try:
            obj = function(file_path)
            if num_trials > 1:
                print("File {0} succesfully accessed after {1} attempts".format(file_path, num_trials))
            num_trials = MAX_READ_TRIALS + 1
        except IOError:
            if os.path.exists(root) and not os.path.exists(file_path):
                raise LisfloodFileError(file_path, error_msg)
            elif num_trials == MAX_READ_TRIALS:
                raise Exception("Cannot access file {0}!\nNetwork down for too long OR bad root directory {1}!".format(file_path, root))
            else:
                num_trials += 1
                print("Trying to access file {0}: attempt n. {1}".format(file_path, num_trials))
                xtime.sleep(READ_PAUSE)
    return obj

def getCalendarType(nc):
    """Get the type of calendar of the open netCDF file passed as argument (http://cfconventions.org/)"""
    try:
        calendar_type = nc.variables["time"].calendar
    except AttributeError:
        calendar_type = "proleptic_gregorian"
        print(LisfloodWarning("""The 'calendar' attribute of the 'time' variable of {} is not set: the default '{}' is used
                              (http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.pdf)""".format(path_nc, calendar_type)))
    return calendar_type


def CalendarInconsistencyWarning(filename, file_calendar, precipitation_calendar):
    return LisfloodWarning("In file {}, time.calendar attribute ({}) differs from that of precipitation ({})!".format(filename, file_calendar, precipitation_calendar))



"""

Copyright 2019 European Union

Licensed under the EUPL, Version 1.2 or as soon they will be approved by the European Commission  subsequent versions of the EUPL (the "Licence");

You may not use this work except in compliance with the Licence.
You may obtain a copy of the Licence at:

https://joinup.ec.europa.eu/sites/default/files/inline-files/EUPL%20v1_2%20EN(1).txt

Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the Licence for the specific language governing permissions and limitations under the Licence.

"""


from lisflood.global_modules.add1 import *


class polder(object):

    """
    # ************************************************************
    # ***** POLDER       *****************************************
    # ************************************************************
    """

    def __init__(self, polder_variable):
        self.var = polder_variable

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------

    def initial(self):
        """ initial part of the polder
         module
        """
        # ************************************************************
        # ***** POLDERS
        # ************************************************************


        if option['simulatePolders']:

            g = 9.81
            # Acceleration due to gravity [m /s^2]
            expPolder = 3 / 2
            # Exponent in weir equation [-]
            PolderSites = loadmap('PolderSites')
            PolderSites = ifthen(
                (defined(PolderSites) & self.var.IsChannel), PolderSites)
            # Get rid of any polders that are not part of the channel network
            # IMPORTANT: current implementation can become unstable with kin.
            # wave!!
            IsPolder = pcraster.boolean(PolderSites)
            # Flag that is boolean(1) for polder sites and boolean(0) otherwise
            PolderTotalCapacity = lookupscalar(
                binding['TabPolderTotalCapacity'], PolderSites)
            # total storage capacity of Polder area [m3]
            PolderArea = lookupscalar(binding['TabPolderArea'], PolderSites)
            # Area of polder [m2]
            PolderOFWidth = lookupscalar(
                binding['TabPolderOFWidth'], PolderSites)
            # Width of polder in-/outflow weir [m]
            PolderBottomLevel = lookupscalar(
                binding['TabPolderBottomLevel'], PolderSites)
            # Bottom level of polder, measured above bottom level of channel
            # [m]
            PolderOpeningTime = lookupscalar(
                binding['TabPolderOpeningTime'], PolderSites)
            # Time step at which each polder is opened (regulated mode)
            PolderReleaseTime = lookupscalar(
                binding['TabPolderReleaseTime'], PolderSites)
            # Time step at which water stored in each polder is released back
            # to the channel (regulated mode)

            PolderOpeningTime = loadmap('PolderOpeningTime')
            PolderReleaseTime = loadmap('PolderReleaseTime')
            UnregulatedFlag = ifthen(
                (PolderOpeningTime == -9999) | (PolderReleaseTime == -9999), pcraster.nominal(1))
            # Flag to indicate all polders that are unregulated
            PolderLevel = PolderInitialLevelValue
            # Initial polder level [m]
            self.var.PolderStorageIniM3 = cover(
                PolderLevel * PolderArea, scalar(0.0))
            # Compute polder storage [m3]
            self.var.PolderStorageM3 = self.var.PolderStorageIniM3
            # Set to initial value
            PolderCapacity = pcraster.max(
                PolderTotalCapacity - PolderStorageM3, scalar(0.0))
            # Remaining storage capacity of polder [m3]

#        else:
            # Set polder storage to zero if no polders are simulated
            # (dummy code)
#            self.var.PolderStorageIniM3 = scalar(0.0)
            # Initial polder storage [m3]
#            self.var.PolderStorageM3 = self.var.PolderStorageIniM3

        # Initialising cumulative output variables
        # These are all needed to compute the cumulative mass balance error

#        self.var.ChannelToPolderM3Dt = globals.inZero.copy()

    @staticmethod
    def dynamic_init():
        """ dynamic part of the polder module
            initialising polders
        """

        # ************************************************************
        # ***** POLDER
        # ************************************************************
        # if option['InitLisflood']==False:    # only with no InitLisflood
        # if option['simulatePolders']:
        # #
        # #
        # ClosedFlag=if(time() lt PolderOpeningTime, nominal(2));
        # FillingFlag=if(time() ge PolderOpeningTime and time() lt PolderReleaseTime, nominal(3));
        # EmptyingFlag=if(time() ge PolderReleaseTime, nominal(4));
        # Flags to indicate that regulated polders are closed/filling/emptying
        # (yields nonsense results for unregulated polders, but see below!)
        # #
        # #            PolderStatus=cover(UnregulatedFlag,ClosedFlag, FillingFlag,EmptyingFlag);
        # Status of polder. PolderStatus can assume the following values:
        # #        		#
        # 1: unregulated (PolderOpeningTime or PolderReleaseTime equal to -9999)
        # 2: closed (time step smaller than PolderOpeningTime)
        # 3: filling (time step greater than or equal to PolderOpeningTime and smaller than PolderReleaseTime)
        # 4: emptying (time step greater than or equal to PolderReleaseTime)
        # #
        # #
        # Water heads in channel (hc) and polder (hp) (i.e. water levels above outflow opening)
        # #
        # hc=if(IsPolder,max(WaterLevelDyn-PolderBottomLevel,0.0),0.0);
        # Head in channel [m]
        # hp=if(IsPolder,PolderLevel,0.0);
        # Head in polder
        # #
        # First compute weir factor
        # #
        # weirFactor=if(hc eq hp,0);
        # Weir factor equals zero if heads are identical in channel and polder (no flow)
        # #
        # weirFactor=if(hc gt hp, sqrt(1-((hp/hc)**16)), weirFactor);
        # hc greater than hp, so flow from channel into polder
        # #
        # weirFactor=if(hp gt hc, sqrt(1-((hc/hp)**16)), weirFactor);
        # gp > hc, so flow from polder back into channel
        # #
        # Compute fluxes from Poleni's modified weir equation
        # #
        # QChannelToPolder=if(hc gt hp,mu*weirFactor*PolderOFWidth*sqrt(2*g)*(hc**expPolder),0.0);
        # QPolderToChannel=if(hp gt hc,mu*weirFactor*PolderOFWidth*sqrt(2*g)*(hp**expPolder),0.0);
        # In- and outgoing fluxes [cu m/s] (both either positive or zero)
        # #
        # Correct fluxes for storage capacity and actual storage
        # #
        # QChannelToPolder=min(QChannelToPolder,PolderCapacity*InvDtSec);
        # No flow into polder beyond its capacity
        # #
        # #        	QPolderToChannel=min(QPolderToChannel, PolderStorageM3*InvDtSec);
        # No flow out of polder once it is empty
        # #
        # Correct fluxes for polder status (in case of regulated polders)
        # #
        # QChannelToPolder=if(PolderStatus eq 2 or PolderStatus eq 4, 0.0, QChannelToPolder);
        # No flow into polder if polder is closed or emptying
        # #
        # QPolderToChannel=if(PolderStatus eq 2 or PolderStatus eq 3, 0.0, QPolderToChannel);
        # No flow out of polder if polder is closed or filling
        # #
        # #        	PolderFlux=cover(QChannelToPolder-QPolderToChannel, 0.0);
        # Polder flux [cu m/s]. Positive for flow from channel to polder, negative for polder to channel
        # #
        # ChannelToPolderM3=PolderFlux*DtSec;
        # Water to (positive) or from (negative) polder [cu m]
        # #
        # #         	PolderStorageM3 += ChannelToPolderM3;
        # New polder storage [m] (remember ChannelToPolderM3 is *negative* while polder is emptying, hence
        # PolderStorageM3 will decrease in that case)
        # #
        # PolderLevel=PolderStorageM3/PolderArea;
        # New polder level [m]
        # #
        # PolderCapacity=max(PolderTotalCapacity-PolderStorageM3,0.0);
        # Remaining storage capacity of polder [cu m]
        # #
        # #
        # QDeltaPolder=(ChannelToPolderM3-ChannelToPolderM3Old)*InvNoRoutSteps;
        # Water to (positive) or from (negative) polder [cu m]
        # #

    @staticmethod
    def dynamic_inloop():
        """ dynamic part of the polder routine
           inside the sub time step routing routine
        """

        # ************************************************************
        # ***** POLDER
        # ************************************************************

        # if option['simulatePolders']:
        # #
        # ChannelToPolderM3Dt=(ChannelToPolderM3Old+(NoRoutingExecuted+1)*QDeltaPolder)*InvNoRoutSteps;
        # flow from (or to) polder per substep
        # #
        pass
